# -*- coding: UTF-8 -*-
# @Time    : 2022/9/14 23:25
# @File    : plot.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import random
from typing import Any
from antgo.pipeline.engine import *
import cv2
import numpy as np


@register
class resize_op(object):
    interp_dict = {
        'NEAREST': cv2.INTER_NEAREST,
        'LINEAR': cv2.INTER_LINEAR,
        'CUBIC': cv2.INTER_CUBIC,
        'AREA': cv2.INTER_AREA,
        'LANCZOS4': cv2.INTER_LANCZOS4
    }
    def __init__(self, out_size, interp='LINEAR') -> None:
        self.resize_w = out_size[0]
        self.resize_h = out_size[1]
        self.interp = interp

    def __call__(self, image):
        image = cv2.resize(
            image, 
            (self.resize_w, self.resize_h), 
            interpolation=self.interp_dict[self.interp])
        return image


@register
class keep_ratio_op(object):
    def __init__(self, aspect_ratio) -> None:
        self.aspect_ratio = aspect_ratio

    def __call__(self, image):
        rhi, rwi = image.shape[:2]
        if abs(rwi / rhi - self.aspect_ratio) > 0.0001:
            if rwi / rhi > self.aspect_ratio:
                nwi = rwi
                nhi = int(rwi / self.aspect_ratio)
            else:
                nhi = rhi
                nwi = int(rhi * self.aspect_ratio)

            # 随机填充
            assert(nhi >= rhi)
            top_padding = (nhi - rhi) // 2
            bottom_padding = (nhi - rhi) - top_padding

            assert(nwi >= rwi)
            left_padding = (nwi - rwi)//2
            right_padding = (nwi - rwi) - (nwi - rwi)//2
  
            # 调整image
            image = cv2.copyMakeBorder(image, top_padding, bottom_padding, left_padding, right_padding,
                                            cv2.BORDER_CONSTANT, value=(128, 128, 128))
            
        return image


@register
class preprocess_op(object):
    def __init__(self, mean, std, permute=None, expand_dim=False) -> None:
        mean = np.array(mean, dtype=np.float32)
        std = np.array(std, dtype=np.float32)
        
        self.mean = np.float64(mean.reshape(1, -1))
        self.stdinv = 1 / np.float64(std.reshape(1, -1))
        self.permute = permute
        self.expand_dim = expand_dim
    
    def __call__(self, image):
        image = image.astype(np.float32)
        cv2.subtract(image, self.mean, image)  # inplace
        cv2.multiply(image, self.stdinv, image)  # inplace
        
        if self.permute is not None:
            image = np.transpose(image, self.permute)
            
        if self.expand_dim:
            image = np.expand_dims(image, 0)
        return image

@register
class center_crop_op(object):
    def __init__(self, size) -> None:
        self.crop_height = size[0]
        self.crop_width = size[1]
    
    def __call__(self, image):
        image_height, image_width = image.shape[:2]
        if self.crop_width > image_width or self.crop_height > image_height:
            padding_ltrb = [
                (self.crop_width - image_width) // 2 if self.crop_width > image_width else 0,
                (self.crop_height - image_height) // 2 if self.crop_height > image_height else 0,
                (self.crop_width - image_width + 1) // 2 if self.crop_width > image_width else 0,
                (self.crop_height - image_height + 1) // 2 if self.crop_height > image_height else 0,
            ]
            pad_left, pad_top, pad_right, pad_bottom = padding_ltrb
            image = np.pad(image, ((pad_top, pad_bottom), (pad_left, pad_right)), 'constant', constant_values=(0, 0))

            image_height, image_width = image.shape[:2]
            if self.crop_width == image_width and self.crop_height == image_height:
                return image
        
        crop_top = int(round((image_height - self.crop_height) / 2.))
        crop_left = int(round((image_width - self.crop_width) / 2.))
        crop_image = image[crop_top:crop_top+self.crop_height, crop_left:crop_left+self.crop_width].copy()
        return crop_image


@register
class nms(object):
    def __init__(self, iou_thres=0.2) -> None:
        self.iou_thres = iou_thres
    
    def __call__(self, boxes, labels) -> Any:
        """ 非极大值抑制 """
        x1 = boxes[:, 0]
        y1 = boxes[:, 1]
        x2 = boxes[:, 2]
        y2 = boxes[:, 3]
        scores = boxes[:, 4]
        areas = (x2-x1) * (y2-y1)
        keep = []

        # 按置信度进行排序
        index = np.argsort(scores)[::-1]

        while(index.size):
            # 置信度最高的框
            i = index[0]
            keep.append(index[0])

            if(index.size == 1): # 如果只剩一个框，直接返回
                break

            # 计算交集左下角与右上角坐标
            inter_x1 = np.maximum(x1[i], x1[index[1:]])
            inter_y1 = np.maximum(y1[i], y1[index[1:]])
            inter_x2 = np.minimum(x2[i], x2[index[1:]])
            inter_y2 = np.minimum(y2[i], y2[index[1:]])
            # 计算交集的面积
            inter_area = np.maximum(inter_x2-inter_x1, 0) * np.maximum(inter_y2-inter_y1, 0)
            # 计算当前框与其余框的iou
            iou = inter_area / (areas[index[1:]] + areas[i] - inter_area)
            ids = np.where(iou < self.iou_thres)[0]
            index = index[ids+1]

        return boxes[keep], labels[keep]


@register
def convert_rgb2bgr_op(x):
    x = cv2.cvtColor(x, cv2.COLOR_BGR2RGB)    
    return x


# -*- coding: UTF-8 -*-
# @Time    : 2022/9/14 23:25
# @File    : plot.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import random
from antgo.pipeline.engine import *
from antgo.utils import colormap
import cv2


def plot_one_box(x, img, color=None, label=None, line_thickness=3):
  # Plots one bounding box on image img
  tl = line_thickness or round(0.002 * (img.shape[0] + img.shape[1]) / 2) + 1  # line/font thickness
  color = color or [random.randint(0, 255) for _ in range(3)]
  c1, c2 = (int(x[0]), int(x[1])), (int(x[2]), int(x[3]))
  cv2.rectangle(img, c1, c2, color, thickness=tl, lineType=cv2.LINE_AA)
  if label:
    tf = max(tl - 1, 1)  # font thickness
    t_size = cv2.getTextSize(label, 0, fontScale=tl / 3, thickness=tf)[0]
    c2 = c1[0] + t_size[0], c1[1] - t_size[1] - 3
    cv2.rectangle(img, c1, c2, color, -1, cv2.LINE_AA)  # filled
    cv2.putText(img, label, (c1[0], c1[1] - 2), 0, tl / 3, [225, 255, 255], thickness=tf, lineType=cv2.LINE_AA)


@register
class plot_bbox(object):
  def __init__(self, thres=0.0, color=None, line_thickness=1, category_map=None, ignore_category=-1):
    self.color = color
    self.line_thickness = line_thickness
    self.thres = thres
    self.category_map = category_map
    self.ignore_category = ignore_category

  def __call__(self, image, detbbox, detlabel=None):
    assert(detbbox.shape[1] == 6 or detbbox.shape[1] == 5)
    if detbbox.shape[1] == 6:
      for (*xyxy, conf, cls), label in zip(detbbox, detlabel):
        if self.ignore_category >= 0:
          if int(cls) == self.ignore_category:
            continue
        
        bbox_label = str(int(cls))
        if label != '':
          bbox_label = label

        if self.category_map is not None:
          bbox_label = self.category_map[bbox_label]
          
        bbox_color = [random.randint(0, 255) for _ in range(3)]
        if self.color is not None:
          bbox_color = self.color[int(cls)]
        plot_one_box(xyxy, image, label=bbox_label, color=bbox_color, line_thickness=self.line_thickness)
    else:
      for (*xyxy, conf), label in zip(detbbox, detlabel):
        if conf < self.thres:
          continue
        if self.ignore_category >= 0:
          if int(label) == self.ignore_category:
            continue
                
        bbox_label = str(int(label))
        if self.category_map is not None:
          bbox_label = self.category_map[bbox_label]
        
        bbox_color = [random.randint(0, 255) for _ in range(3)]
        if self.color is not None:
          bbox_color = self.color[int(label)]
        plot_one_box(xyxy, image, label=bbox_label, color=bbox_color, line_thickness=self.line_thickness)


    return image
  

@register
class plot_text(object):
  def __init__(self, label_map=None, color=None, pos=None, font_scale=1, thickness=1):
    self.pos = pos
    self.color = color
    self.thickness = thickness
    self.font_scale = font_scale
    self.label_map = label_map
  
  def __call__(self, image, text):
    pos = self.pos
    if pos is None:
      pos = (int(image.shape[1]/2), int(image.shape[0]/2))
    
    color = self.color
    if color is None:
      color = (0, 0, 255)
    
    if not isinstance(text, str):
      text = self.label_map[int(text)]
    image = cv2.putText(image, text, pos, cv2.FONT_HERSHEY_COMPLEX, self.font_scale, color, self.thickness)
    return image


# 优化器配置
optimizer = dict(type='SGD', lr=0.05,  weight_decay=5e-4, momentum=0.9, nesterov=True)
optimizer_config = dict(grad_clip=None)

# 学习率调度配置
lr_config = dict(
    policy='CosineAnnealing',
    min_lr=1e-5,
    warmup_by_epoch=False,
    warmup_iters=2000,
    warmup='linear'    
)

# 日志配置
log_config = dict(
    interval=10,    
    hooks=[
        dict(type='TextLoggerHook'),
    ])

# 模型配置
model = dict(
    type='EncoderDecoder',
    backbone=dict(
        type='DDRKetNetF',
        in_channels=3
    ),
    decode_head=dict(
        type='SimpleHead',
        in_channels=32,
        channels=32,
        out_channels=21,
        ignore_index=255
    )
)

# 描述模型基本信息
info = dict(
    backbone=dict(
        channels=[96,128,160],
        shapes=[32,16,8]
    )
)

# checkpoint配置
checkpoint_config = dict(interval=1, out_dir='./output/')       

# 数据配置
data=dict(
    train=dict(
        type='Pascal2012',
        train_or_test='train',
        dir='./pascal2012_dataset',
        ext_params=dict(task_type='SEGMENTATION', aug=True),
        pipeline=[
            dict(type='Rotation'),
            dict(type='ResizeS', target_dim=(256,256)),
            dict(type='ColorDistort'),
            dict(type='RandomFlipImage', swap_ids=[], swap_labels=[]),
            dict(type='ToTensor', keys=['image']),
            dict(type='Normalize', mean=(0.491400, 0.482158, 0.4465231), std=(0.247032, 0.243485, 0.2615877), keys=['image']),
            dict(type='UnSqueeze', axis=0, keys=['segments'])
        ],
        inputs_def=dict(
            fields=['image', 'segments'],
        )
    ),
    train_dataloader=dict(
        samples_per_gpu=128, 
        workers_per_gpu=2,
        drop_last=True,
        shuffle=True,
    ),
    val=dict(
        type='Pascal2012',
        train_or_test='val',
        dir='./pascal2012_dataset',
        ext_params=dict(task_type='SEGMENTATION', aug=True),
        pipeline=[
            dict(type='ResizeS', target_dim=(256,256)),
            dict(type='ToTensor', keys=['image']),
            dict(type='Normalize', mean=(0.491400, 0.482158, 0.4465231), std=(0.247032, 0.243485, 0.2615877), keys=['image']),
            dict(type='UnSqueeze', axis=0, keys=['segments'])
        ],        
        inputs_def=dict(
            fields=['image', 'segments'],
        )
    ),
    val_dataloader=dict(
        samples_per_gpu=128, 
        workers_per_gpu=2,
        drop_last=False,
        shuffle=False,
    ),
    test=dict(
        type='Pascal2012',
        train_or_test='val',
        dir='./pascal2012_dataset',
        ext_params=dict(task_type='SEGMENTATION', aug=True),
        pipeline=[
            dict(type='ResizeS', target_dim=(256,256)),
            dict(type='ToTensor', keys=['image']),
            dict(type='Normalize', mean=(0.491400, 0.482158, 0.4465231), std=(0.247032, 0.243485, 0.2615877), keys=['image']),
            dict(type='UnSqueeze', axis=0, keys=['segments'])
        ],                
        inputs_def=dict(
            fields=['image', 'segments'],
        )
    ),
    test_dataloader=dict(
        samples_per_gpu=128, 
        workers_per_gpu=2,
        drop_last=False,
        shuffle=False,
    )
)

# 评估方案配置
evaluation=dict(out_dir='./output/', interval=1, metric=dict(type='SegMIOU', class_num=21), save_best='miou', rule='greater')

# 导出配置
export=dict(
    input_shape_list = [[1,3,256,256]],
    input_name_list=["image"],
    output_name_list=["heatmap"]
)

max_epochs = 600


# -*- coding: UTF-8 -*-
# @Time    : 2022/5/1 14:36
# @File    : test_torch_c.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
# import torch
# import torch.nn as nn
# import antgo.net.torch2paddle.nn as nn
import antgo.framework.torch2paddle.mapper as torch
import antgo.framework.torch2paddle.mapper.nn as nn
import antgo.framework.torch2paddle.mapper.optimizer as optim
import time
import os
from MobileNetV2 import *
import argparse
from read_ImageNetData import ImageNetData



def train_model(args, model, criterion, optimizer, scheduler, num_epochs, dataset_sizes):
  since = time.time()
  resumed = False

  best_model_wts = model.state_dict()

  for epoch in range(args.start_epoch + 1, num_epochs):

    # Each epoch has a training and validation phase
    for phase in ['train', 'val']:
      if phase == 'train':
        if args.start_epoch > 0 and (not resumed):
          scheduler.step(args.start_epoch + 1)
          resumed = True
        else:
          scheduler.step(epoch)
        model.train(True)  # Set model to training mode
      else:
        model.train(False)  # Set model to evaluate mode

      running_loss = 0.0
      running_corrects = 0

      tic_batch = time.time()
      # Iterate over data.
      for i, (inputs, labels) in enumerate(dataloders[phase]):
        # wrap them in Variable
        # if use_gpu:
        #   inputs = Variable(inputs.cuda())
        #   labels = Variable(labels.cuda())
        # else:
        #   inputs, labels = Variable(inputs), Variable(labels)

        # zero the parameter gradients
        optimizer.zero_grad()

        # forward
        outputs = model(inputs)
        preds = torch.max(outputs.data, 1)
        loss = criterion(outputs, labels)

        # backward + optimize only if in training phase
        if phase == 'train':
          loss.backward()
          optimizer.step()

        # statistics
        running_loss += loss.data[0]
        running_corrects += torch.sum(preds == labels.data)

        batch_loss = running_loss / ((i + 1) * args.batch_size)
        batch_acc = running_corrects / ((i + 1) * args.batch_size)

        if phase == 'train' and i % args.print_freq == 0:
          print('[Epoch {}/{}]-[batch:{}/{}] lr:{:.4f} {} Loss: {:.6f}  Acc: {:.4f}  Time: {:.4f}batch/sec'.format(
            epoch, num_epochs - 1, i, round(dataset_sizes[phase] / args.batch_size) - 1, scheduler.get_lr()[0], phase,
            batch_loss, batch_acc.item(), \
                   args.print_freq / (time.time() - tic_batch)))
          tic_batch = time.time()

        break

      epoch_loss = running_loss / dataset_sizes[phase]
      epoch_acc = running_corrects / dataset_sizes[phase]

      print('{} Loss: {:.4f} Acc: {:.4f}'.format(
        phase, epoch_loss, epoch_acc.item()))

      break

    if (epoch + 1) % args.save_epoch_freq == 0:
      if not os.path.exists(args.save_path):
        os.makedirs(args.save_path)
      torch.save(model, os.path.join(args.save_path, "epoch_" + str(epoch) + ".pth.tar"))

    break

  time_elapsed = time.time() - since
  print('Training complete in {:.0f}m {:.0f}s'.format(
    time_elapsed // 60, time_elapsed % 60))

  # load best model weights
  model.load_state_dict(best_model_wts)
  return model


if __name__ == '__main__':

  parser = argparse.ArgumentParser(description="PyTorch implementation of SENet")
  parser.add_argument('--data-dir', type=str, default="/ImageNet")
  parser.add_argument('--batch-size', type=int, default=16)
  parser.add_argument('--num-class', type=int, default=1000)
  parser.add_argument('--num-epochs', type=int, default=100)
  parser.add_argument('--lr', type=float, default=0.045)
  parser.add_argument('--num-workers', type=int, default=0)
  parser.add_argument('--gpus', type=str, default=0)
  parser.add_argument('--print-freq', type=int, default=10)
  parser.add_argument('--save-epoch-freq', type=int, default=1)
  parser.add_argument('--save-path', type=str, default="output")
  parser.add_argument('--resume', type=str, default="", help="For training from one checkpoint")
  parser.add_argument('--start-epoch', type=int, default=0, help="Corresponding to the epoch of resume ")
  args = parser.parse_args()

  # read data
  dataloders, dataset_sizes = ImageNetData(args)

  # use gpu or not
  use_gpu = torch.cuda.is_available()
  print("use_gpu:{}".format(use_gpu))

  # get model
  model = mobilenetv2_19(num_classes=args.num_class)

  if args.resume:
    if os.path.isfile(args.resume):
      print(("=> loading checkpoint '{}'".format(args.resume)))
      checkpoint = torch.load(args.resume)
      base_dict = {'.'.join(k.split('.')[1:]): v for k, v in list(checkpoint.state_dict().items())}
      model.load_state_dict(base_dict)
    else:
      print(("=> no checkpoint found at '{}'".format(args.resume)))

  if use_gpu:
    model = model.cuda()
    model = torch.nn.DataParallel(model, device_ids=[int(i) for i in args.gpus.strip().split(',')])

  # define loss function
  criterion = nn.CrossEntropyLoss()

  # Observe that all parameters are being optimized
  optimizer_ft = optim.SGD(model.parameters(), lr=args.lr, momentum=0.9, weight_decay=0.00004)

  # Decay LR by a factor of 0.1 every 7 epochs
  exp_lr_scheduler = optim.lr_scheduler.StepLR(optimizer_ft, step_size=1, gamma=0.98)

  model = train_model(args=args,
                      model=model,
                      criterion=criterion,
                      optimizer=optimizer_ft,
                      scheduler=exp_lr_scheduler,
                      num_epochs=args.num_epochs,
                      dataset_sizes=dataset_sizes)



import os
import logging
from aligo import Aligo

def share_data_func(args):
    data_name = args.name
    shared_data_info_folder = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'resource', 'dataset')
    
    support_info_list = []
    for file_name in os.listdir(shared_data_info_folder):
        if file_name.startswith('vgg-face2'):
            continue
        support_info_list.append(file_name.split('.')[0])
    support_info_list.append('vgg-face2')

    if data_name is None:
        logging.error('--name must be set')
        print(f'share data support {support_info_list}')
        return
    
    if data_name not in support_info_list:
        print(f'share data only support {support_info_list}')
        return
    
    # 需要用户权限认证
    ali = Aligo()
    file_obj = ali.get_folder_by_path('/dataset/')
    if file_obj is None:
        ali.create_folder('/dataset')
    
    # 特殊处理vgg-face2
    if data_name == 'vgg-face2':
        with open(os.path.join(shared_data_info_folder, f'{data_name}-data.txt'), 'r') as fp:
            data_content = fp.read()
        with open(os.path.join(shared_data_info_folder, f'{data_name}-meta.txt'), 'r') as fp:
            meta_content = fp.read()
        
        file_obj = ali.get_folder_by_path(f'/dataset/{data_name}')
        if file_obj is None:
            ali.create_folder(f'/dataset/{data_name}')
        
        file_obj = ali.get_folder_by_path(f'/dataset/{data_name}/data')
        if file_obj is None:
            ali.create_folder(f'/dataset/{data_name}/data')
            file_obj = ali.get_folder_by_path(f'/dataset/{data_name}/data')
        
        ali.save_files_by_aligo(data_content, file_obj.file_id)
            
        file_obj = ali.get_folder_by_path(f'/dataset/{data_name}/meta')
        if file_obj is None:
            ali.create_folder(f'/dataset/{data_name}/meta')  
            file_obj = ali.get_folder_by_path(f'/dataset/{data_name}/meta')

        ali.save_files_by_aligo(meta_content, file_obj.file_id)      
        return
    
    with open(os.path.join(shared_data_info_folder, f'{data_name}.txt'), 'r') as fp:
        content = fp.read()
    
    file_obj = ali.get_folder_by_path(f'/dataset/{data_name}')
    if file_obj is None:
        ali.create_folder(f'/dataset/{data_name}')
        file_obj = ali.get_folder_by_path(f'/dataset/{data_name}')

    ali.save_files_by_aligo(content, file_obj.file_id)


# share_data_func('vgg-face2')


from typing import Dict
from antgo.framework.helper.models.detectors.base import BaseDetector
from antgo.framework.helper.base_module import BaseModule


class MultiSteamModule(BaseModule):
    def __init__(
        self, model=dict(), train_cfg=None, test_cfg=None, init_cfg=None):
        super(MultiSteamModule, self).__init__()
        self.submodules = list(model.keys())
        for k, v in model.items():
            setattr(self, k, v)

        self.train_cfg = train_cfg
        self.test_cfg = test_cfg
        self.inference_on = self.test_cfg.get("inference_on", None)

    def model(self, **kwargs):
        if "submodule" in kwargs:
            assert (
                kwargs["submodule"] in self.submodules
            ), "Detector does not contain submodule {}".format(kwargs["submodule"])
            model = getattr(self, kwargs["submodule"])
        else:
            model = getattr(self, self.inference_on)
        return model

    def freeze(self, model_ref: str):
        assert model_ref in self.submodules
        model = getattr(self, model_ref)
        model.eval()
        for param in model.parameters():
            param.requires_grad = False

    def forward_test(self, image, image_meta=None, **kwargs):
        kwargs.update({
            'image_meta': image_meta
        })
        return self.model(**kwargs).forward_test(image, **kwargs)

    def extract_feat(self, image):
        return self.model().extract_feat(image)

    def simple_test(self, image, image_meta=None, **kwargs):
        kwargs.update({
            'image_meta': image_meta
        })        
        return self.model(**kwargs).simple_test(image, **kwargs)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
import torch
import torch.nn as nn
import torch.nn.functional as F
from ...builder import DISTILL_LOSSES



'''
DML with only two networks
'''

@DISTILL_LOSSES.register_module()
class DML(nn.Module):
	'''
	Deep Mutual Learning
	https://zpascal.net/cvpr2018/Zhang_Deep_Mutual_Learning_CVPR_2018_paper.pdf
	'''
	def __init__(self):
		super(DML, self).__init__()

	def forward(self, out1, out2):
		loss = F.kl_div(F.log_softmax(out1, dim=1),
						F.softmax(out2, dim=1),
						reduction='batchmean')

		return loss


# Once for All: Train One Network and Specialize it for Efficient Deployment
# Han Cai, Chuang Gan, Tianzhe Wang, Zhekai Zhang, Song Han
# International Conference on Learning Representations (ICLR), 2020.

import os
import json
import numpy as np
from tqdm import tqdm
import torch
import torch.utils.data

from antgo.framework.helper.models.nas.utils import list_mean

__all__ = ["net_setting2id", "net_id2setting", "AccuracyDataset"]


def net_setting2id(net_setting):
    return json.dumps(net_setting)


def net_id2setting(net_id):
    return json.loads(net_id)


class RegDataset(torch.utils.data.Dataset):
    def __init__(self, inputs, targets):
        super(RegDataset, self).__init__()
        self.inputs = inputs
        self.targets = targets

    def __getitem__(self, index):
        return self.inputs[index], self.targets[index]

    def __len__(self):
        return self.inputs.size(0)


class AccuracyDataset:
    def __init__(self, path):
        self.path = path
        os.makedirs(self.path, exist_ok=True)

    @property
    def net_id_path(self):
        return os.path.join(self.path, "net_id.dict")

    @property
    def acc_src_folder(self):
        return os.path.join(self.path, "src")

    @property
    def acc_dict_path(self):
        return os.path.join(self.path, "acc.dict")

    # TODO: support parallel building
    def build_acc_dataset(
        self, run_manager, ofa_network, n_arch=1000, image_size_list=None
    ):
        # load net_id_list, random sample if not exist
        if os.path.isfile(self.net_id_path):
            net_id_list = json.load(open(self.net_id_path))
        else:
            net_id_list = set()
            while len(net_id_list) < n_arch:
                net_setting = ofa_network.sample_active_subnet()
                net_id = net_setting2id(net_setting)
                net_id_list.add(net_id)
            net_id_list = list(net_id_list)
            net_id_list.sort()
            json.dump(net_id_list, open(self.net_id_path, "w"), indent=4)

        image_size_list = (
            [128, 160, 192, 224] if image_size_list is None else image_size_list
        )

        with tqdm(
            total=len(net_id_list) * len(image_size_list), desc="Building Acc Dataset"
        ) as t:
            for image_size in image_size_list:
                # load val dataset into memory
                val_dataset = []
                run_manager.run_config.data_provider.assign_active_img_size(image_size)
                for images, labels in run_manager.run_config.valid_loader:
                    val_dataset.append((images, labels))
                # save path
                os.makedirs(self.acc_src_folder, exist_ok=True)
                acc_save_path = os.path.join(
                    self.acc_src_folder, "%d.dict" % image_size
                )
                acc_dict = {}
                # load existing acc dict
                if os.path.isfile(acc_save_path):
                    existing_acc_dict = json.load(open(acc_save_path, "r"))
                else:
                    existing_acc_dict = {}
                for net_id in net_id_list:
                    net_setting = net_id2setting(net_id)
                    key = net_setting2id({**net_setting, "image_size": image_size})
                    if key in existing_acc_dict:
                        acc_dict[key] = existing_acc_dict[key]
                        t.set_postfix(
                            {
                                "net_id": net_id,
                                "image_size": image_size,
                                "info_val": acc_dict[key],
                                "status": "loading",
                            }
                        )
                        t.update()
                        continue
                    ofa_network.set_active_subnet(**net_setting)
                    run_manager.reset_running_statistics(ofa_network)
                    net_setting_str = ",".join(
                        [
                            "%s_%s"
                            % (
                                key,
                                "%.1f" % list_mean(val)
                                if isinstance(val, list)
                                else val,
                            )
                            for key, val in net_setting.items()
                        ]
                    )
                    loss, (top1, top5) = run_manager.validate(
                        run_str=net_setting_str,
                        net=ofa_network,
                        data_loader=val_dataset,
                        no_logs=True,
                    )
                    info_val = top1

                    t.set_postfix(
                        {
                            "net_id": net_id,
                            "image_size": image_size,
                            "info_val": info_val,
                        }
                    )
                    t.update()

                    acc_dict.update({key: info_val})
                    json.dump(acc_dict, open(acc_save_path, "w"), indent=4)

    def merge_acc_dataset(self, image_size_list=None):
        # load existing data
        merged_acc_dict = {}
        for fname in os.listdir(self.acc_src_folder):
            if ".dict" not in fname:
                continue
            image_size = int(fname.split(".dict")[0])
            if image_size_list is not None and image_size not in image_size_list:
                print("Skip ", fname)
                continue
            full_path = os.path.join(self.acc_src_folder, fname)
            partial_acc_dict = json.load(open(full_path))
            merged_acc_dict.update(partial_acc_dict)
            print("loaded %s" % full_path)
        json.dump(merged_acc_dict, open(self.acc_dict_path, "w"), indent=4)
        return merged_acc_dict

    def build_acc_data_loader(
        self, arch_encoder, n_training_sample=None, batch_size=256, n_workers=16
    ):
        # load data
        acc_dict = json.load(open(self.acc_dict_path))
        X_all = []
        Y_all = []
        with tqdm(total=len(acc_dict), desc="Loading data") as t:
            for k, v in acc_dict.items():
                dic = json.loads(k)
                X_all.append(arch_encoder.arch2feature(dic))
                Y_all.append(v / 100.0)  # range: 0 - 1
                t.update()
        base_acc = np.mean(Y_all)
        # convert to torch tensor
        X_all = torch.tensor(X_all, dtype=torch.float)
        Y_all = torch.tensor(Y_all)

        # random shuffle
        shuffle_idx = torch.randperm(len(X_all))
        X_all = X_all[shuffle_idx]
        Y_all = Y_all[shuffle_idx]

        # split data
        idx = X_all.size(0) // 5 * 4 if n_training_sample is None else n_training_sample
        val_idx = X_all.size(0) // 5 * 4
        X_train, Y_train = X_all[:idx], Y_all[:idx]
        X_test, Y_test = X_all[val_idx:], Y_all[val_idx:]
        print("Train Size: %d," % len(X_train), "Valid Size: %d" % len(X_test))

        # build data loader
        train_dataset = RegDataset(X_train, Y_train)
        val_dataset = RegDataset(X_test, Y_test)

        train_loader = torch.utils.data.DataLoader(
            train_dataset,
            batch_size=batch_size,
            shuffle=True,
            pin_memory=False,
            num_workers=n_workers,
        )
        valid_loader = torch.utils.data.DataLoader(
            val_dataset,
            batch_size=batch_size,
            shuffle=False,
            pin_memory=False,
            num_workers=n_workers,
        )

        return train_loader, valid_loader, base_acc



# -*- coding: UTF-8 -*-
# @Time    : 2022/9/12 14:18
# @File    : __init__.py.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

from .io import image_decode, image_save, image_base64_decode, image_download
from .plot import plot_bbox, plot_text
from .process import resize_op, keep_ratio_op, preprocess_op


__all__ = [
    'image_decode', 'image_save','plot_bbox', 'image_base64_decode', 'plot_text', 'image_download', 'resize_op', 'keep_ratio_op', 'preprocess_op'
]



# -*- coding: UTF-8 -*-
# @Time    : 2022/4/27 20:38
# @File    : vis.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

import os
import os.path as osp
import cv2
import numpy as np
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import matplotlib as mpl
import matplotlib.patches as patches
from PIL import Image, ImageDraw


def get_keypoint_rgb(skeleton):
    rgb_dict= {}
    for joint_id in range(len(skeleton)):
        joint_name = skeleton[joint_id]['name']

        if joint_name.endswith('thumb_null'):
            rgb_dict[joint_name] = (255, 0, 0)
        elif joint_name.endswith('thumb3'):
            rgb_dict[joint_name] = (255, 51, 51)
        elif joint_name.endswith('thumb2'):
            rgb_dict[joint_name] = (255, 102, 102)
        elif joint_name.endswith('thumb1'):
            rgb_dict[joint_name] = (255, 153, 153)
        elif joint_name.endswith('thumb0'):
            rgb_dict[joint_name] = (255, 204, 204)
        elif joint_name.endswith('index_null'):
            rgb_dict[joint_name] = (0, 255, 0)
        elif joint_name.endswith('index3'):
            rgb_dict[joint_name] = (51, 255, 51)
        elif joint_name.endswith('index2'):
            rgb_dict[joint_name] = (102, 255, 102)
        elif joint_name.endswith('index1'):
            rgb_dict[joint_name] = (153, 255, 153)
        elif joint_name.endswith('middle_null'):
            rgb_dict[joint_name] = (255, 128, 0)
        elif joint_name.endswith('middle3'):
            rgb_dict[joint_name] = (255, 153, 51)
        elif joint_name.endswith('middle2'):
            rgb_dict[joint_name] = (255, 178, 102)
        elif joint_name.endswith('middle1'):
            rgb_dict[joint_name] = (255, 204, 153)
        elif joint_name.endswith('ring_null'):
            rgb_dict[joint_name] = (0, 128, 255)
        elif joint_name.endswith('ring3'):
            rgb_dict[joint_name] = (51, 153, 255)
        elif joint_name.endswith('ring2'):
            rgb_dict[joint_name] = (102, 178, 255)
        elif joint_name.endswith('ring1'):
            rgb_dict[joint_name] = (153, 204, 255)
        elif joint_name.endswith('pinky_null'):
            rgb_dict[joint_name] = (255, 0, 255)
        elif joint_name.endswith('pinky3'):
            rgb_dict[joint_name] = (255, 51, 255)
        elif joint_name.endswith('pinky2'):
            rgb_dict[joint_name] = (255, 102, 255)
        elif joint_name.endswith('pinky1'):
            rgb_dict[joint_name] = (255, 153, 255)
        else:
            rgb_dict[joint_name] = (230, 230, 0)
        
    return rgb_dict

def vis_keypoints(img, kps, score, skeleton, filename, score_thr=0.4, line_width=3, circle_rad = 3, save_path='./'):
    rgb_dict = get_keypoint_rgb(skeleton)
    _img = Image.fromarray(img.transpose(1,2,0).astype('uint8')) 
    draw = ImageDraw.Draw(_img)
    for i in range(len(skeleton)):
        joint_name = skeleton[i]['name']
        pid = skeleton[i]['parent_id']
        parent_joint_name = skeleton[pid]['name']
        
        kps_i = (kps[i][0].astype(np.int32), kps[i][1].astype(np.int32))
        kps_pid = (kps[pid][0].astype(np.int32), kps[pid][1].astype(np.int32))

        if score[i] > score_thr and score[pid] > score_thr and pid != -1:
            draw.line([(kps[i][0], kps[i][1]), (kps[pid][0], kps[pid][1])], fill=rgb_dict[parent_joint_name], width=line_width)
        if score[i] > score_thr:
            draw.ellipse((kps[i][0]-circle_rad, kps[i][1]-circle_rad, kps[i][0]+circle_rad, kps[i][1]+circle_rad), fill=rgb_dict[joint_name])
        if score[pid] > score_thr and pid != -1:
            draw.ellipse((kps[pid][0]-circle_rad, kps[pid][1]-circle_rad, kps[pid][0]+circle_rad, kps[pid][1]+circle_rad), fill=rgb_dict[parent_joint_name])
    
    _img.save(osp.join(save_path, filename))


def __render_2d_xyz(kps_3d,score, skeleton, ax, rgb_dict, score_thr=0.4, axis='x'):
    kps_3d = kps_3d.copy()

    kps_2d = np.zeros((kps_3d.shape[0], 2))
    a_name = 'x'
    b_name = 'y'
    if axis == 'x':
        # show y,z
        kps_2d[:,0] = kps_3d[:,2]
        kps_2d[:,1] = -kps_3d[:,1]
        a_name = 'z'
        b_name = 'y'
    elif axis == 'y':
        # show x,z
        kps_2d[:,0] = kps_3d[:,0]
        kps_2d[:,1] = kps_3d[:,2]
        a_name = 'x'
        b_name = 'z'
    else:
        # show x,y
        kps_2d[:,0] = kps_3d[:,0]
        kps_2d[:,1] = -kps_3d[:,1]
        a_name = 'x'
        b_name = 'y'
    ax.set_title(f'show {a_name}-{b_name} plane')
    ax.set_xlabel(a_name)
    ax.set_ylabel(b_name)    
    for i in range(len(skeleton)):
        joint_name = skeleton[i]['name']
        pid = skeleton[i]['parent_id']
        parent_joint_name = skeleton[pid]['name']

        a = np.array([kps_2d[i,0], kps_2d[pid,0]])
        b = np.array([kps_2d[i,1], kps_2d[pid,1]])

        if score[i] > score_thr and score[pid] > score_thr and pid != -1:
            ax.plot(a,b, c = np.array(rgb_dict[parent_joint_name])/255., linewidth = 3)

        if score[i] > score_thr:
            ax.scatter(kps_2d[i,0], kps_2d[i,1], c = np.array(rgb_dict[joint_name]).reshape(1,3)/255., marker='o')
        if score[pid] > score_thr and pid != -1:
            ax.scatter(kps_2d[pid,0], kps_2d[pid,1], c = np.array(rgb_dict[parent_joint_name]).reshape(1,3)/255., marker='o')


def vis_3d_keypoints(kps_3d, score, skeleton, filename, score_thr=0.4, line_width=3, circle_rad=3, save_path='./'):
    fig = plt.figure()
    fig.subplots_adjust(wspace=0.3, hspace=0.4)
    ax = fig.add_subplot(221, projection='3d')
    rgb_dict = get_keypoint_rgb(skeleton)
    
    for i in range(len(skeleton)):
        joint_name = skeleton[i]['name']
        pid = skeleton[i]['parent_id']
        parent_joint_name = skeleton[pid]['name']

        x = np.array([kps_3d[i,0], kps_3d[pid,0]])
        y = np.array([kps_3d[i,1], kps_3d[pid,1]])
        z = np.array([kps_3d[i,2], kps_3d[pid,2]])

        if score[i] > score_thr and score[pid] > score_thr and pid != -1:
            ax.plot(x, z, -y, c = np.array(rgb_dict[parent_joint_name])/255., linewidth = line_width)
        if score[i] > score_thr:
            ax.scatter(kps_3d[i,0], kps_3d[i,2], -kps_3d[i,1], c = np.array(rgb_dict[joint_name]).reshape(1,3)/255., marker='o')
        if score[pid] > score_thr and pid != -1:
            ax.scatter(kps_3d[pid,0], kps_3d[pid,2], -kps_3d[pid,1], c = np.array(rgb_dict[parent_joint_name]).reshape(1,3)/255., marker='o')

    # x 方向
    ax = fig.add_subplot(222)
    __render_2d_xyz(kps_3d,score, skeleton, ax, rgb_dict,score_thr, axis='x')

    # y
    ax = fig.add_subplot(223)
    __render_2d_xyz(kps_3d,score, skeleton, ax, rgb_dict,score_thr, axis='y')

    # z
    ax = fig.add_subplot(224)
    __render_2d_xyz(kps_3d,score, skeleton, ax, rgb_dict,score_thr, axis='z')
    
    fig.savefig(osp.join(save_path, filename), dpi=fig.dpi)
    plt.close()


def vis_2d_boxes_in_image(image, boxes, labels, save_path='./'):
    color_map = [(255,0,0),(0,0,255),(0,255,0),(255,255,0),(0,255,255),(255,0,255),(0,0,0),(255,255,255)]
    for box, label in zip(boxes, labels):
        x1,y1,x2,y2 = box
        
        color_index = int(label) % len(color_map)
        cv2.rectangle(image, (int(x1),int(y1)), (int(x2), int(y2)), color_map[color_index], 2)
    cv2.imwrite(save_path, image)


def vis_3d_boxes_in_image(image, boxes, camera_param, camera_model):
    pass

def vis_3d_points_in_image(image, points, skeleton, camera_param, camera_model='fisheye'):
    
    pass

def vis_2d_points_in_image(image, points, skeleton):
    pass


def vis_mano_in_image(image, pose, shape, trans, camera_param, camera_model='fisheye'):
    pass



def imshow(image, bboxes=None, keypoints=None):
    # Create figure and axes
    fig, ax = plt.subplots(1)

    # Display the image
    image = image.copy()
    if keypoints is not None:
        for k in range(keypoints.shape[0]):
            x, y = keypoints[k]
            image[int(y), int(x), :] = 255

    ax.imshow(image)

    if bboxes is not None:
        for b in range(bboxes.shape[0]):
            x0, y0, x1, y1 = bboxes[b]
            width = x1 - x0
            height = y1 - y0
            # Create a Rectangle patch
            rect = patches.Rectangle((x0, y0), width, height, linewidth=1, edgecolor='r', facecolor='none')
            # Add the patch to the Axes
            ax.add_patch(rect)

    plt.show()
    plt.waitforbuttonpress()


# -*- coding: UTF-8 -*-
# @Time    : 2022/9/18 12:48
# @File    : runas_op.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from typing import Callable

from antgo.pipeline.engine import register

# pylint: disable=import-outside-toplevel
# pylint: disable=invalid-name

@register(name='runas_op')
class runas_op:
    """
    Convert a user-defined function as an operator and execute.

    Args:
        func (`Callable`):
            The user-defined function.

    Examples:

    >>> from antgo.pipeline.functional import DataCollection
    >>> from antgo.pipeline.functional.entity import Entity
    >>> entities = [Entity(a=i, b=i) for i in range(5)]
    >>> dc = DataCollection(entities)
    >>> res = dc.runas_op['a', 'b'](func=lambda x: x - 1).to_list()
    >>> res[0].a == res[0].b + 1
    True
    """
    def __init__(self, func: Callable):
        self._func = func

    def __call__(self, *args, **kws):
        return self._func(*args, **kws)


# __test__ = {'run': run.__doc__}



# Copyright (c) OpenMMLab. All rights reserved.
from .base import LoggerHook
from .neptune import NeptuneLoggerHook
from .text import TextLoggerHook

__all__ = [
    'LoggerHook', 'TextLoggerHook', 'NeptuneLoggerHook'
]



import torch
from antgo.framework.helper.runner import BaseModule
from abc import ABCMeta, abstractmethod

class BaseClassifier(BaseModule, metaclass=ABCMeta):
    """Base class for classifiers."""

    def __init__(self, init_cfg=None):
        super(BaseClassifier, self).__init__(init_cfg)

    @property
    def with_neck(self):
        return hasattr(self, 'neck') and self.neck is not None

    @property
    def with_head(self):
        return hasattr(self, 'head') and self.head is not None

    def forward_test(self, image, **kwargs):
        """
        Args:
            image (List[Tensor]): the outer list indicates test-time
                augmentations and inner Tensor should have a shape NxCxHxW,
                which contains all images in the batch.
        """
        if isinstance(image, torch.Tensor):
            image = [image]
        for var, name in [(image, 'imgs')]:
            if not isinstance(var, list):
                raise TypeError(f'{name} must be a list, but got {type(var)}')

        if len(image) == 1:
            return self.simple_test(image[0], **kwargs)
        else:
            raise NotImplementedError('aug_test has not been implemented')



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
import torch
import torch.nn as nn
import torch.nn.functional as F
from ...builder import DISTILL_LOSSES


@DISTILL_LOSSES.register_module()
class SP(nn.Module):
	'''
	Similarity-Preserving Knowledge Distillation
	https://arxiv.org/pdf/1907.09682.pdf
	'''
	def __init__(self):
		super(SP, self).__init__()

	def forward(self, fm_s, fm_t):
		fm_s = fm_s.view(fm_s.size(0), -1)
		G_s  = torch.mm(fm_s, fm_s.t())
		norm_G_s = F.normalize(G_s, p=2, dim=1)

		fm_t = fm_t.view(fm_t.size(0), -1)
		G_t  = torch.mm(fm_t, fm_t.t())
		norm_G_t = F.normalize(G_t, p=2, dim=1)

		loss = F.mse_loss(norm_G_s, norm_G_t)

		return loss


# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import os
import cv2
import numpy as np
import random
import math

def load_img(path, order='RGB'):
    
    # load
    img = cv2.imread(path, cv2.IMREAD_COLOR | cv2.IMREAD_IGNORE_ORIENTATION)
    if not isinstance(img, np.ndarray):
        raise IOError("Fail to read %s" % path)

    if order=='RGB':
        img = img[:,:,::-1].copy()
    
    img = img.astype(np.float32)
    return img

def load_skeleton(path, joint_num):

    # load joint info (name, parent_id)
    skeleton = [{} for _ in range(joint_num)]
    with open(path) as fp:
        for line in fp:
            if line[0] == '#': continue
            splitted = line.split(' ')
            joint_name, joint_id, joint_parent_id = splitted
            joint_id, joint_parent_id = int(joint_id), int(joint_parent_id)
            skeleton[joint_id]['name'] = joint_name
            skeleton[joint_id]['parent_id'] = joint_parent_id
    # save child_id
    for i in range(len(skeleton)):
        joint_child_id = []
        for j in range(len(skeleton)):
            if skeleton[j]['parent_id'] == i:
                joint_child_id.append(j)
        skeleton[i]['child_id'] = joint_child_id
    
    return skeleton

def get_aug_config():
    trans_factor = 0.08
    scale_factor = 0.15
    rot_factor = 60
    color_factor = 0.1
    
    trans = [
        np.random.uniform(-trans_factor, trans_factor), 
        np.random.uniform(-trans_factor, trans_factor)
    ]
    scale = np.clip(np.random.randn(), -1.0, 1.0) * scale_factor + 1.0
    # rot = np.clip(np.random.randn(), -2.0,
    #               2.0) * rot_factor if random.random() <= 0.6 else 0
    rot = (np.random.random() * 2 - 1) * rot_factor if random.random() <= 0.6 else 0
    do_flip = random.random() <= 0.5
    c_up = 1.0 + color_factor
    c_low = 1.0 - color_factor
    # 仅考虑灰度图情况
    color_scale = random.uniform(c_low, c_up)
    color_scale = np.array([color_scale, color_scale, color_scale])

    do_noise = np.random.random() <= 0.5
    # do_noise = False
    return trans, scale, rot, do_flip, color_scale, do_noise

def augmentation(
    img, bbox, joint_coord, joint_valid, hand_type, 
    mode, joint_type, input_img_shape, allow_geometric_aug=True, allow_flip=True, img_process_callback=None):
    img = img.copy(); 
    joint_coord = joint_coord.copy()
    hand_type = hand_type.copy()

    original_img_shape = img.shape
    joint_num = len(joint_coord)
    
    if mode == 'train':
        trans, scale, rot, do_flip, color_scale, do_noise = get_aug_config()
    else:
        trans, scale, rot, do_flip, color_scale, do_noise = [0,0], 1.0, 0.0, False, np.array([1,1,1]), False
    
    # TODO，考虑非刚性变形，以帮助不同手型的扩展

    if not allow_geometric_aug:
        trans, scale, rot, do_flip = [0,0], 1.0, 0.0, False
    if not allow_flip:
        do_flip = False

    bbox[0] = bbox[0] + bbox[2] * trans[0]
    bbox[1] = bbox[1] + bbox[3] * trans[1]
    # 生成局部patch
    img, trans, inv_trans = generate_patch_image(img, bbox, do_flip, scale, rot, input_img_shape)

    if img_process_callback is not None:
        # 对局部图像进行定制化变换
        img = img_process_callback(img)

    # 颜色尺度变换
    img = np.clip(img * color_scale[None,None,:], 0, 255)
    
    if do_noise:
        # 加噪声
        img = img.astype(np.float)
        noise_delta = (np.random.random(img.shape) * 2 - 1) * 10
        img = img + noise_delta
        img = np.clip(img, 0, 255)
        img = img.astype(np.uint8)
    
    if do_flip:
        # 加翻转
        joint_coord[:,0] = original_img_shape[1] - joint_coord[:,0] - 1
        # 左右手坐标切换
        joint_coord[joint_type['right']], joint_coord[joint_type['left']] = joint_coord[joint_type['left']].copy(), joint_coord[joint_type['right']].copy()
        # 左右手可见性切换
        joint_valid[joint_type['right']], joint_valid[joint_type['left']] = joint_valid[joint_type['left']].copy(), joint_valid[joint_type['right']].copy()
        # 左右手标签切换
        hand_type[0], hand_type[1] = hand_type[1].copy(), hand_type[0].copy()

    # 需要重新设置关键点的可见性
    for i in range(joint_num):
        joint_coord[i,:2] = trans_point2d(joint_coord[i,:2], trans)
        joint_valid[i] = joint_valid[i] * (joint_coord[i,0] >= 0) * (joint_coord[i,0] < input_img_shape[1]) * (joint_coord[i,1] >= 0) * (joint_coord[i,1] < input_img_shape[0])

    return img, joint_coord, joint_valid, hand_type, inv_trans

def transform_input_to_output_space(joint_coord, joint_valid, rel_root_depth, root_valid, root_joint_idx, joint_type, input_img_shape, output_hm_shape, bbox_3d_size, output_root_hm_shape, bbox_3d_size_root):
    # transform to output heatmap space
    joint_coord = joint_coord.copy(); joint_valid = joint_valid.copy()
    
    joint_coord[:,0] = joint_coord[:,0] / input_img_shape[1] * output_hm_shape[2]
    joint_coord[:,1] = joint_coord[:,1] / input_img_shape[0] * output_hm_shape[1]
    joint_coord[joint_type['right'],2] = joint_coord[joint_type['right'],2] - joint_coord[root_joint_idx['right'],2]
    joint_coord[joint_type['left'],2] = joint_coord[joint_type['left'],2] - joint_coord[root_joint_idx['left'],2]
  
    joint_coord[:,2] = (joint_coord[:,2] / (bbox_3d_size/2) + 1)/2. * output_hm_shape[0]
    joint_valid = joint_valid * ((joint_coord[:,2] >= 0) * (joint_coord[:,2] < output_hm_shape[0])).astype(np.float32)
    rel_root_depth = (rel_root_depth / (bbox_3d_size_root/2) + 1)/2. * output_root_hm_shape
    root_valid = root_valid * ((rel_root_depth >= 0) * (rel_root_depth < output_root_hm_shape)).astype(np.float32)
    
    return joint_coord, joint_valid, rel_root_depth, root_valid

def get_bbox(joint_img, joint_valid):
    x_img = joint_img[:,0][joint_valid==1]; y_img = joint_img[:,1][joint_valid==1];
    xmin = min(x_img); ymin = min(y_img); xmax = max(x_img); ymax = max(y_img);

    x_center = (xmin+xmax)/2.; width = xmax-xmin;
    xmin = x_center - 0.5*width*1.2
    xmax = x_center + 0.5*width*1.2
    
    y_center = (ymin+ymax)/2.; height = ymax-ymin;
    ymin = y_center - 0.5*height*1.2
    ymax = y_center + 0.5*height*1.2

    bbox = np.array([xmin, ymin, xmax-xmin, ymax-ymin]).astype(np.float32)
    return bbox

def process_bbox(bbox, original_img_shape, input_img_shape):

    # aspect ratio preserving bbox
    w = bbox[2]
    h = bbox[3]
    c_x = bbox[0] + w/2.
    c_y = bbox[1] + h/2.
    aspect_ratio = input_img_shape[1]/input_img_shape[0]
    if w > aspect_ratio * h:
        h = w / aspect_ratio
    elif w < aspect_ratio * h:
        w = h * aspect_ratio
    bbox[2] = w*1.25
    bbox[3] = h*1.25
    bbox[0] = c_x - bbox[2]/2.
    bbox[1] = c_y - bbox[3]/2.

    return bbox

def generate_patch_image(cvimg, bbox, do_flip, scale, rot, out_shape):
    img = cvimg.copy()
    img_height, img_width, img_channels = img.shape

    bb_c_x = float(bbox[0] + 0.5*bbox[2])
    bb_c_y = float(bbox[1] + 0.5*bbox[3])
    bb_width = float(bbox[2])
    bb_height = float(bbox[3])

    if do_flip:
        img = img[:, ::-1, :]
        bb_c_x = img_width - bb_c_x - 1
    
    trans = gen_trans_from_patch_cv(bb_c_x, bb_c_y, bb_width, bb_height, out_shape[1], out_shape[0], scale, rot)
    img_patch = cv2.warpAffine(img, trans, (int(out_shape[1]), int(out_shape[0])), flags=cv2.INTER_LINEAR)
    img_patch = img_patch.astype(np.float32)
    inv_trans = gen_trans_from_patch_cv(bb_c_x, bb_c_y, bb_width, bb_height, out_shape[1], out_shape[0], scale, rot, inv=True)

    return img_patch, trans, inv_trans

def rotate_2d(pt_2d, rot_rad):
    x = pt_2d[0]
    y = pt_2d[1]
    sn, cs = np.sin(rot_rad), np.cos(rot_rad)
    xx = x * cs - y * sn
    yy = x * sn + y * cs
    return np.array([xx, yy], dtype=np.float32)

def gen_trans_from_patch_cv(c_x, c_y, src_width, src_height, dst_width, dst_height, scale, rot, inv=False):
    # augment size with scale
    src_w = src_width * scale
    src_h = src_height * scale
    src_center = np.array([c_x, c_y], dtype=np.float32)

    # augment rotation
    rot_rad = np.pi * rot / 180
    src_downdir = rotate_2d(np.array([0, src_h * 0.5], dtype=np.float32), rot_rad)
    src_rightdir = rotate_2d(np.array([src_w * 0.5, 0], dtype=np.float32), rot_rad)

    dst_w = dst_width
    dst_h = dst_height
    dst_center = np.array([dst_w * 0.5, dst_h * 0.5], dtype=np.float32)
    dst_downdir = np.array([0, dst_h * 0.5], dtype=np.float32)
    dst_rightdir = np.array([dst_w * 0.5, 0], dtype=np.float32)

    src = np.zeros((3, 2), dtype=np.float32)
    src[0, :] = src_center
    src[1, :] = src_center + src_downdir
    src[2, :] = src_center + src_rightdir

    dst = np.zeros((3, 2), dtype=np.float32)
    dst[0, :] = dst_center
    dst[1, :] = dst_center + dst_downdir
    dst[2, :] = dst_center + dst_rightdir
    
    if inv:
        trans = cv2.getAffineTransform(np.float32(dst), np.float32(src))
    else:
        trans = cv2.getAffineTransform(np.float32(src), np.float32(dst))

    trans = trans.astype(np.float32)
    return trans

def trans_point2d(pt_2d, trans):
    src_pt = np.array([pt_2d[0], pt_2d[1], 1.]).T
    dst_pt = np.dot(trans, src_pt)
    return dst_pt[0:2]

def cam2pixel(cam_coord, f, c):
    x = cam_coord[:, 0] / (cam_coord[:, 2] + 1e-8) * f[0] + c[0]
    y = cam_coord[:, 1] / (cam_coord[:, 2] + 1e-8) * f[1] + c[1]
    z = cam_coord[:, 2]
    img_coord = np.concatenate((x[:,None], y[:,None], z[:,None]),1)
    return img_coord

def pixel2cam(pixel_coord, f, c):
    x = (pixel_coord[:, 0] - c[0]) / f[0] * pixel_coord[:, 2]
    y = (pixel_coord[:, 1] - c[1]) / f[1] * pixel_coord[:, 2]
    z = pixel_coord[:, 2]
    cam_coord = np.concatenate((x[:,None], y[:,None], z[:,None]),1)
    return cam_coord

def cam2pixel_omnidir(p, cam0d_i, cam0d_mu, cam0d_d):
    point = p.copy()
    point = point.reshape(p.shape[0],1,3)
    uv,_ = cv2.omnidir.projectPoints(point, np.zeros(3), np.zeros(3), cam0d_i, cam0d_mu, cam0d_d)

    return uv.squeeze(1)

def pixel2cam_omnidir(uv,z,K,D,xi):
    '''
    2.5d到3d投影
    :param uv:  图片 uv坐标
    :param z: 深度
    :param K: 相机内参
    :param D: 相机畸变
    :param xi: omni相机 xi参数
    :return: 3d坐标
    '''
    u = (uv[:, 0] - K[0, 2]) / K[0, 0]
    v = (uv[:, 1] - K[1, 2]) / K[1, 1]
    _u = u.copy()
    _v = v.copy()

    for i in range(20):
        r2 = u ** 2 + v ** 2
        r4 = r2 * r2
        u = (_u - 2 * D[2] * u * v - D[3] * (r2 + 2 * u * u)) / (1 + D[0] * r2 + D[1] * r4)
        v = (_v - 2 * D[3] * u * v - D[2] * (r2 + 2 * v * v)) / (1 + D[0] * r2 + D[1] * r4)

    r2 = u ** 2 + v ** 2
    a = r2 + 1
    b = 2 * xi * r2
    cc = r2 * xi * xi - 1
    Zs = (-b + np.sqrt(b * b - 4 * a * cc)) / (2 * a)
    xw = np.vstack([u * (Zs + xi), v * (Zs + xi), Zs])
    xyz = xw*(z/xw[2])
    xyz = np.transpose(xyz)
    return xyz

def pixel2cam_invK(pixel_coord, inv_K):
    z = pixel_coord[:,2:]
    uvones = pixel_coord.copy()    
    uvones[:,2] = 1.0
    test_tt = np.matmul(inv_K, np.transpose(uvones))
    test_tt = np.transpose(test_tt)
    test_tt[:,2:] = z
    return test_tt

def pixel2fisheyecam(pixel_coord, cam_param):
    pass

def fisheyecam2pixel(cam_coord, cam_param):
    pass

def world2cam(world_coord, R, T):
    cam_coord = np.dot(R, world_coord - T)
    return cam_coord

def cam2world(cam_coord, inv_R, T):
    world_coord = np.dot(inv_R, cam_coord) + T
    return world_coord


# Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import copy
import functools
import collections
import traceback
import numpy as np
from antgo.utils import logger
from antgo.utils.parallel_map import ParallelMap

class Compose(object):
    def __init__(self, transforms, ctx=None):
        self.transforms = transforms
        self.ctx = ctx

    def __call__(self, data):
        ctx = self.ctx if self.ctx else {}
        for f in self.transforms:
            try:
                data = f(data, ctx)
            except Exception as e:
                stack_info = traceback.format_exc()
                logger.warn("fail to map op [{}] with error: {} and stack:\n{}".
                            format(f, e, str(stack_info)))
                raise e
        return data


def _calc_img_weights(roidbs):
    """ calculate the probabilities of each sample
    """
    imgs_cls = []
    num_per_cls = {}
    img_weights = []
    for i, roidb in enumerate(roidbs):
        img_cls = set([k for cls in roidbs[i]['gt_class'] for k in cls])
        imgs_cls.append(img_cls)
        for c in img_cls:
            if c not in num_per_cls:
                num_per_cls[c] = 1
            else:
                num_per_cls[c] += 1

    for i in range(len(roidbs)):
        weights = 0
        for c in imgs_cls[i]:
            weights += 1 / num_per_cls[c]
        img_weights.append(weights)
    # probabilities sum to 1
    img_weights = img_weights / np.sum(img_weights)
    return img_weights


def _has_empty(item):
    def empty(x):
        if isinstance(x, np.ndarray) and x.size == 0:
            return True
        elif isinstance(x, collections.Sequence) and len(x) == 0:
            return True
        else:
            return False

    if isinstance(item, collections.Sequence) and len(item) == 0:
        return True
    if item is None:
        return True
    if empty(item):
        return True
    return False


def _segm(samples):
    assert 'gt_poly' in samples
    segms = samples['gt_poly']
    if 'is_crowd' in samples:
        is_crowd = samples['is_crowd']
        if len(segms) != 0:
            assert len(segms) == is_crowd.shape[0]

    gt_masks = []
    valid = True
    for i in range(len(segms)):
        segm = segms[i]
        gt_segm = []
        if 'is_crowd' in samples and is_crowd[i]:
            gt_segm.append([[0, 0]])
        else:
            for poly in segm:
                if len(poly) == 0:
                    valid = False
                    break
                gt_segm.append(np.array(poly).reshape(-1, 2))
        if (not valid) or len(gt_segm) == 0:
            break
        gt_masks.append(gt_segm)
    return gt_masks


def batch_arrange(batch_samples, fields):
    def im_shape(samples, dim=3):
        # hard code
        assert 'h' in samples
        assert 'w' in samples
        if dim == 3:  # RCNN, ..
            return np.array((samples['h'], samples['w'], 1), dtype=np.float32)
        else:  # YOLOv3, ..
            return np.array((samples['h'], samples['w']), dtype=np.int32)

    arrange_batch = []
    for samples in batch_samples:
        one_ins = ()
        for i, field in enumerate(fields):
            if field == 'gt_mask':
                one_ins += (_segm(samples), )
            elif field == 'im_shape':
                one_ins += (im_shape(samples), )
            elif field == 'im_size':
                one_ins += (im_shape(samples, 2), )
            else:
                if field == 'is_difficult':
                    field = 'difficult'
                assert field in samples, '{} not in samples'.format(field)
                one_ins += (samples[field], )
        arrange_batch.append(one_ins)
    return arrange_batch


class Reader(object):
    """
    Args:
        dataset (DataSet): DataSet object
        sample_transforms (list of BaseOperator): a list of sample transforms
            operators.
        batch_transforms (list of BaseOperator): a list of batch transforms
            operators.
        batch_size (int): batch size.
        shuffle (bool): whether shuffle dataset or not. Default False.
        drop_last (bool): whether drop last batch or not. Default False.
        drop_empty (bool): whether drop sample when it's gt is empty or not.
            Default True.
        mixup_epoch (int): mixup epoc number. Default is -1, meaning
            not use mixup.
        cutmix_epoch (int): cutmix epoc number. Default is -1, meaning
            not use cutmix.
        class_aware_sampling (bool): whether use class-aware sampling or not.
            Default False.
        worker_num (int): number of working threads/processes.
            Default -1, meaning not use multi-threads/multi-processes.
        use_process (bool): whether use multi-processes or not.
            It only works when worker_num > 1. Default False.
        bufsize (int): buffer size for multi-threads/multi-processes,
            please note, one instance in buffer is one batch data.
        memsize (str): size of shared memory used in result queue when
            use_process is true. Default 3G.
        inputs_def (dict): network input definition use to get input fields,
            which is used to determine the order of returned data.
        devices_num (int): number of devices.
    """

    def __init__(self,
                 dataset=None,
                 sample_transforms=None,
                 batch_transforms=None,
                 batch_size=None,
                 shuffle=False,
                 drop_last=False,
                 drop_empty=True,
                 mixup_epoch=-1,
                 cutmix_epoch=-1,
                 class_aware_sampling=False,
                 worker_num=-1,
                 use_process=False,
                 num_classes=80,
                 bufsize=-1,
                 memsize='3G',
                 inputs_def=None,
                 devices_num=1):
        self._dataset = dataset
        self._fields = copy.deepcopy(inputs_def[
            'fields']) if inputs_def else None

        # transform
        self._sample_transforms = Compose(sample_transforms,
                                          {'fields': self._fields})
        self._batch_transforms = None
        if batch_transforms:
            self._batch_transforms = Compose(batch_transforms,
                                             {'fields': self._fields})

        # # data
        # if inputs_def and inputs_def.get('multi_scale', False):
        #     from ppdet.modeling.architectures.input_helper import multiscale_def
        #     im_shape = inputs_def[
        #         'image_shape'] if 'image_shape' in inputs_def else [
        #             3, None, None
        #         ]
        #     _, ms_fields = multiscale_def(im_shape, inputs_def['num_scales'],
        #                                   inputs_def['use_flip'])
        #     self._fields += ms_fields

        self._batch_size = batch_size
        self._shuffle = shuffle
        self._drop_last = drop_last
        self._drop_empty = drop_empty

        # sampling
        self._mixup_epoch = mixup_epoch
        self._cutmix_epoch = cutmix_epoch
        self._class_aware_sampling = class_aware_sampling

        self._load_img = False
        self._sample_num = self._dataset.size

        if self._class_aware_sampling:
            self.img_weights = np.ones((self._sample_num), dtype=np.float32) / self._sample_num

        self._indexes = None
        self._pos = -1
        self._epoch = -1
        self._curr_iter = 0

        # multi-process
        self._worker_num = worker_num
        self._parallel = None
        if self._worker_num > -1:
            task = functools.partial(self.worker, self._drop_empty)
            bufsize = devices_num * 2 if bufsize == -1 else bufsize
            self._parallel = ParallelMap(self, task, worker_num, bufsize,
                                         use_process, memsize)

    def __call__(self):
        if self._worker_num > -1:
            return self._parallel
        else:
            return self

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def iterator_value(self):
        obj = None
        if self._worker_num > -1:
            obj = self._parallel
        else:
            obj = self

        for _batch in obj:
            if len(_batch) > 0:
                yield _batch

        # reset random
        self.reset()

    def reset(self):
        """implementation of Dataset.reset
        """
        if self._epoch < 0:
            self._epoch = 0
        else:
            self._epoch += 1

        # 样本索引
        self.indexes = [i for i in range(self.size())]
        # 重采样
        if self._class_aware_sampling:
            self.indexes = np.random.choice(
                self._sample_num,
                self._sample_num,
                replace=True,
                p=self.img_weights)

        # 洗牌
        if self._shuffle:
            trainer_id = int(os.getenv("PADDLE_TRAINER_ID", 0))
            np.random.seed(self._epoch + trainer_id)
            np.random.shuffle(self.indexes)

        # 混合
        if self._mixup_epoch > 0 and len(self.indexes) < 2:
            logger.debug("Disable mixup for dataset samples "
                         "less than 2 samples")
            self._mixup_epoch = -1
        if self._cutmix_epoch > 0 and len(self.indexes) < 2:
            logger.info("Disable cutmix for dataset samples "
                        "less than 2 samples")
            self._cutmix_epoch = -1

        self._pos = 0

    def next(self):
        if self._epoch < 0:
            self.reset()
        if self.drained():
            raise StopIteration
        batch = self._load_batch()
        self._curr_iter += 1
        if self._drop_last and len(batch) < self._batch_size:
            raise StopIteration
        if self._worker_num > -1:
            return batch
        else:
            return self.worker(self._drop_empty, batch)

    def _load_batch(self):
        batch = []
        bs = 0
        while bs != self._batch_size:
            if self._pos >= self.size():
                break
            
            pos = self.indexes[self._pos]
            sample = self._dataset.sample(pos)
            if sample is None:
                # None，需要直接跳过
                continue

            sample["curr_iter"] = self._curr_iter
            self._pos += 1

            if self._drop_empty and self._fields and 'gt_mask' in self._fields:
                if _has_empty(_segm(sample)):
                    #logger.warn('gt_mask is empty or not valid in {}'.format(
                    #    sample['im_file']))
                    continue
            if self._drop_empty and self._fields and 'gt_bbox' in self._fields:
                if _has_empty(sample['gt_bbox']):
                    #logger.warn('gt_bbox {} is empty or not valid in {}, '
                    #   'drop this sample'.format(
                    #    sample['im_file'], sample['gt_bbox']))
                    continue

            if self._load_img and 'im_file' in sample:
                sample['image'] = self._load_image(sample['im_file'])

            if self._epoch < self._mixup_epoch:
                num = len(self.indexes)
                mix_idx = np.random.randint(1, num)
                mix_idx = self.indexes[(mix_idx + self._pos - 1) % num]

                sample['mixup'] = self._dataset.sample(mix_idx)
                sample['mixup']["curr_iter"] = self._curr_iter
                if self._load_img and 'im_file' in sample['mixup']:
                    sample['mixup']['image'] = self._load_image(sample['mixup'][
                        'im_file'])
            if self._epoch < self._cutmix_epoch:
                num = len(self.indexes)
                mix_idx = np.random.randint(1, num)
                sample['cutmix'] = self._dataset.sample(mix_idx)
                sample['cutmix']["curr_iter"] = self._curr_iter
                if self._load_img and 'im_file' in sample['cutmix']:
                    sample['cutmix']['image'] = self._load_image(sample[
                        'cutmix']['im_file'])

            batch.append(sample)
            bs += 1
        return batch

    def worker(self, drop_empty=True, batch_samples=None):
        """
        sample transform and batch transform.
        """
        batch = []
        for sample in batch_samples:
            sample = self._sample_transforms(sample)
            if drop_empty and 'gt_bbox' in sample:
                if _has_empty(sample['gt_bbox']):
                    #logger.warn('gt_bbox {} is empty or not valid in {}, '
                    #   'drop this sample'.format(
                    #    sample['im_file'], sample['gt_bbox']))
                    continue
            batch.append(sample)
            
        if len(batch) > 0 and self._batch_transforms:
            batch = self._batch_transforms(batch)
        if len(batch) > 0 and self._fields:
            batch = batch_arrange(batch, self._fields)
        return batch

    def _load_image(self, filename):
        with open(filename, 'rb') as f:
            return f.read()

    def size(self):
        """ implementation of Dataset.size
        """
        return self._sample_num

    def drained(self):
        """ implementation of Dataset.drained
        """
        assert self._epoch >= 0, 'The first epoch has not begin!'
        return self._pos >= self.size()

    def stop(self):
        if self._parallel:
            self._parallel.stop()


def create_reader(cfg):
    """
    Return iterable data reader.
    """
    reader = Reader(**cfg)()

    def _reader():
        for _batch in reader:
            if len(_batch) > 0:
                yield _batch

        # reset random
        reader.reset()
        
    return _reader



# -*- coding: UTF-8 -*-
# @Time    : 2020/10/26 10:26 上午
# @File    : download.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.utils import logger
from bs4 import BeautifulSoup
import os
import sys
import cv2
import numpy as nps
import time
import threading
import requests
import re
import uuid
import os
try:
    import queue
except:
    import Queue as queue


def baidu_download(keyward, download_params, save_dir, process_queue=None):
    headers = {
        'Accept-Language': 'zh-CN,zh;q=0.8,zh-TW;q=0.7,zh-HK;q=0.5,en-US;q=0.3,en;q=0.2',
        'Connection': 'keep-alive',
        'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64; rv:60.0) Gecko/20100101 Firefox/60.0',
        'Upgrade-Insecure-Requests': '1'
    }
    A = requests.Session()
    A.headers = headers
    url = 'https://image.baidu.com/search/flip?tn=baiduimage&ie=utf-8&word=' + keyward + '&pn='

    def __baidu_find_and_download(waiting_process_queue, search_url, session, dir, max_page_num=50):
        t = 0
        num = 0
        while t < max_page_num:
            Url = search_url + str(t)
            t = t+1
            try:
                Result = session.get(Url, timeout=7, allow_redirects=False)
            except BaseException:
                t = t + 60
                continue
            else:
                pic_url = re.findall('"objURL":"(.*?)",', Result.text, re.S)  # 先利用正则表达式找到图片url
                for each in pic_url:
                    logger.info("Downloading(%d) %s."%(num+1, str(each)))
                    try:
                        if each is not None:
                            pic = requests.get(each, timeout=7)
                        else:
                            continue
                    except BaseException:
                        logger.error("Couldnt download %s."%each)
                        continue
                    else:
                        # 分配唯一文件标识
                        file_folder = os.path.join(dir, 'test')
                        if not os.path.exists(file_folder):
                            os.makedirs(file_folder)
                        
                        file_path = os.path.join(file_folder, 'baidu_%s.jpg'%str(uuid.uuid4()))
                        with open(file_path, 'wb') as fp:
                            fp.write(pic.content)
                        num += 1

                        logger.info("Finish download %s ."% str(each))
                        # 加入等待处理队列
                        if waiting_process_queue is not None:
                            waiting_process_queue.put(file_path)

        # 添加结束标记
        if waiting_process_queue is not None:
            waiting_process_queue.put(None)

    # 搜索和下载
    if process_queue is not None:
        t = threading.Thread(target=__baidu_find_and_download, args=(process_queue, url, A, save_dir))
        t.start()
    else:
        __baidu_find_and_download(process_queue, url, A, save_dir)


def bing_download(keyward, download_params, save_dir, process_queue=None):
    headers = {
        'Accept-Language': 'zh-CN,zh;q=0.8,zh-TW;q=0.7,zh-HK;q=0.5,en-US;q=0.3,en;q=0.2',
        'Connection': 'keep-alive',
        'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64; rv:60.0) Gecko/20100101 Firefox/60.0',
        'Upgrade-Insecure-Requests': '1'
    }
    A = requests.Session()
    A.headers = headers
    #url = 'https://image.baidu.com/search/flip?tn=baiduimage&ie=utf-8&word=' + keyward + '&pn='
    url = 'http://cn.bing.com/images/async?q='+keyward+'&first={0}&count=35&relp=35&lostate=r&mmasync=1&dgState=x*175_y*848_h*199_c*1_i*106_r*0'

    def __bing_find_and_download(waiting_process_queue, search_url, session, dir, max_page_num=50):
        t = 0
        num = 0
        while t < max_page_num:
            Url = search_url.format(t*35+1)

            t = t+1
            try:
                Result = session.get(Url, timeout=7, allow_redirects=False)
            except BaseException:
                t = t + 60
                continue
            else:
                pic_url = re.findall('src="(.*?)"', Result.text, re.S)  # 先利用正则表达式找到图片url
                for each in pic_url:
                    logger.info("Downloading(%d) %s."%(num+1, str(each)))
                    try:
                        if each is not None:
                            pic = requests.get(each, timeout=7)
                        else:
                            continue
                    except BaseException:
                        logger.error("Couldnt download %s."%each)
                        continue
                    else:
                        # 分配唯一文件标识
                        file_folder = os.path.join(dir, 'test')
                        if not os.path.exists(file_folder):
                            os.makedirs(file_folder)
                        
                        file_path = os.path.join(file_folder, 'bing_%s.jpg'%str(uuid.uuid4()))
                        with open(file_path, 'wb') as fp:
                            fp.write(pic.content)
                        num += 1
                        logger.info("Finish download %s ."% str(each))

                        # 加入等待处理队列
                        if waiting_process_queue is not None:
                            waiting_process_queue.put(file_path)
        
        # 结束标记
        if waiting_process_queue is not None:
            waiting_process_queue.put(None)

    # 搜索和下载
    if process_queue is not None:
        t = threading.Thread(target=__bing_find_and_download, args=(process_queue, url, A, save_dir))
        t.start()
    else:
        __bing_find_and_download(process_queue, url, A, save_dir)


def google_download(keyward, download_params, save_dir, process_queue=None):
    pass


def vcg_download(keyword, download_params, download_save_dir, process_queue=None):
    # 视觉中国
    headers = {
        'Accept-Language': 'zh-CN,zh;q=0.8,zh-TW;q=0.7,zh-HK;q=0.5,en-US;q=0.3,en;q=0.2',
        'Connection': 'keep-alive',
        'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64; rv:60.0) Gecko/20100101 Firefox/60.0',
        'Upgrade-Insecure-Requests': '1'
    }
    download_data_type = 'image'
    if download_params is not None:
        download_data_type = download_params.get('download_data_type', 'image')
        if download_data_type not in ['image','video']:
            logger.warn('Only support donwload data_type: image or video')
            download_data_type = 'image'

    def __vcg_img_download(waiting_process_queue, save_dir, img_url, keyword, count):
        try:
            logger.info("Downloading(%d) %s."%(count+1, img_url))
            pic = requests.get(img_url, timeout=7)
        except BaseException:
            logger.error("Couldnt download %s."%img_url)
            return
        else:
            file_prefix = 'VCG_'+keyword + '_' + str(count)
            file_name = file_prefix + '.jpg' if download_data_type == 'image' else file_prefix + '.mp4'
            file_path = os.path.join(save_dir,  file_name)
            fp = open(file_path, 'wb')
            fp.write(pic.content)
            fp.close()
            logger.info("Finish download %s ."%img_url)
            
            if waiting_process_queue is not None:
                waiting_process_queue.put(file_path)

    def __vcg_try_get(A, page_url, try_times=5):
        while try_times != 0:
            Result = A.get(page_url, timeout=7, allow_redirects=True)
            if Result.status_code == 200:
                return Result

            logger.warn("sleep 10s, continue try.")
            time.sleep(10)
            try_times -= 1
        
        return None

    def __vcg_find_and_download(waiting_process_queue, save_dir):
        A = requests.Session()
        A.headers = headers
        query_url = 'https://www.vcg.com/creative-image/{}'.format(keyword)
        if download_data_type == 'video':
            query_url = 'https://www.vcg.com/creative-video-search/{}'.format(keyword)

        download_count = 0
        content = None
        nav_page_list = []
        Result = A.get(query_url, timeout=7, allow_redirects=True)
        if Result.status_code != 200:
            logger.error("%s couldnt connect."%query_url)
            return

        content = Result.text
        soup = BeautifulSoup(content)
        # 1.step 分析待检索页面
        if download_data_type == 'image':
            pages = soup.findAll("a", class_="paginationClickBox")
            last_page = pages[-1]

            total_page = soup.findAll("span", class_="paginationTotal")
            try:
                total_page_num = int(last_page.text)
                if len(total_page) > 0:
                    total_page_num = (int)(total_page[0].text[1:-1])
            except:
                logger.error("Fail to parse nav page, use default 2.")
                total_page_num = 2

            page_nav_url = last_page.get('href').split('?')[0]
            for i in range(2, total_page_num+1):
                nav_page_list.append('%s?page=%d'%(page_nav_url, i))
        else:
            pages = soup.findAll("a", class_="_2IlL4")
            last_page = pages[-2]
            try:
                total_page_num = int(last_page.text)
            except:
                logger.error("Fail to parse nav page, use default 2.")
                total_page_num = 2

            page_nav_url = last_page.get('href').split('?')[0]
            for i in range(2, total_page_num+1):
                nav_page_list.append('%s?page=%d'%(page_nav_url, i))

        # 2.step 分析当前页面的图像
        logger.info("Analyze nav page(%d/%d) %s"%(1, len(nav_page_list), query_url))
        img_url_list = []
        if download_data_type == 'image':
            # 分析图像资源
            img_list = soup.findAll("img", class_="lazyload_hk")
            img_url_list = ["http://{}".format(p.get("data-src")[2:]) for p in img_list]
        else:
            # 分析视频资源
            img_list = soup.findAll('source', type="image/webp")
            img_url_list = []
            video_name_list = []
            for p in img_list:
                kk = p.get('data-srcset')[2:].split('/')[3].split('?')[0].split('_')[0]
                if kk not in video_name_list:
                    video_name_list.append(kk)
                    img_url_list.append('http://gossv.cfp.cn/videos/mts_videos/medium/temp/{}.mp4'.format(kk))

        for img_url in img_url_list:
            __vcg_img_download(waiting_process_queue, save_dir, img_url, keyword, download_count)
            download_count += 1

        # 3.step 继续分析所有等待导航页面
        for page_index, page_url in enumerate(nav_page_list):
            logger.info("Analyze nav page(%d/%d) %s"%(page_index+2, len(nav_page_list), page_url))
            Result = A.get(page_url, timeout=7, allow_redirects=True)
            if Result.status_code != 200:
                logger.warn("Couldnt connect and analyze %s. (page %s)"%((Result.text, page_url)))
                Result = __vcg_try_get(A, page_url, 5)
                if Result is None:
                    logger.warn("Couldnt connect %s, return."%page_url)
                    return

            content = Result.text
            soup = BeautifulSoup(content)

            img_url_list = []
            if download_data_type == 'image':
                # 分析图像资源
                img_list = soup.findAll("img", class_="lazyload_hk")
                img_url_list = ["http://{}".format(p.get("data-src")[2:]) for p in img_list]
            else:
                # 分析视频资源
                img_list = soup.findAll('source', type="image/webp")
                img_url_list = []
                video_name_list = []
                for p in img_list:
                    kk = p.get('data-srcset')[2:].split('/')[3].split('?')[0].split('_')[0]
                    if kk not in video_name_list:
                        video_name_list.append(kk)
                        img_url_list.append('http://gossv.cfp.cn/videos/mts_videos/medium/temp/{}.mp4'.format(kk))

            logger.info("Finding %d data"%len(img_url_list))
            for img_url in img_url_list:
                __vcg_img_download(waiting_process_queue, save_dir, img_url, keyword, download_count)
                download_count += 1

        # 添加结束标记
        if waiting_process_queue is not None:
            waiting_process_queue.put(None)

    # 搜索和下载
    if not os.path.exists(os.path.join(download_save_dir, 'test')):
        os.makedirs(os.path.join(download_save_dir, 'test'))

    if process_queue is not None:
        t = threading.Thread(target=__vcg_find_and_download, args=(process_queue, os.path.join(download_save_dir, 'test')))
        t.start()
    else:
        __vcg_find_and_download(process_queue, os.path.join(download_save_dir, 'test'))





#encoding=utf-8
from __future__ import division
from __future__ import unicode_literals

import numpy as np
import unittest

def apk(actual, predicted, k=10):
    """
    Computes the average precision at k. (only for Information Retrieval Task)
    This function computes the average prescision at k between two lists of
    items.
    References: http://nlp.stanford.edu/IR-book/resource/htmledition/evaluation-of-ranked-retrieval-results-1.resource
    Parameters
    ----------
    actual : list
             A list of elements that are to be predicted (order doesn't matter)
    predicted : list
             A list of predicted elements (order does matter)
    k : int, optional
        The maximum number of predicted elements
    Returns
    -------
    score : double
            The average precision at k over the input lists

    """
    if len(predicted)>k:
        predicted = predicted[:k]

    score = 0.0
    num_hits = 0.0

    for i,p in enumerate(predicted):
        if p in actual and p not in predicted[:i]:
            num_hits += 1.0
            score += num_hits / (i+1.0)

    if not actual:
        return 0.0

    return score / min(len(actual), k)

def s_apk(actual,predicated_s,k=10):
    '''
    Computes the average precision at k. (only for Information Retrieval Task)
    This function computes the average prescision at k between two lists of
    items.
    References: http://nlp.stanford.edu/IR-book/resource/htmledition/evaluation-of-ranked-retrieval-results-1.resource

    Parameters
    ----------
    actual: list
            ground truth label list
    predicated_s: list
            predicated score list
    k: int
            The maximum number of predicted elements
    Returns
    -------
    score : double
            he average precision at k over the input lists
    '''
    if type(actual) == list:
        actual = np.array(actual)
    if type(predicated_s) == list:
        predicated_s = np.array(predicated_s)

    if actual.dtype != int:
        actual = actual.astype(int)

    #absorb some error
    if np.max(actual) > 1:
        actual[np.where(actual > 1)[0]] = 1
    if len(predicated_s.shape) > 1 and predicated_s.shape[1] > 1:
        predicated_s = predicated_s[:, 0]
    if len(predicated_s.shape) > 1:
        predicated_s = predicated_s.flatten()

    positive_actual = np.where(actual == 1)[0]
    sorted_predicated_s = sorted(zip(predicated_s,range(predicated_s.shape[0])),reverse=True)
    sorted_predicated = [n for m,n in sorted_predicated_s]

    return apk(positive_actual.tolist(),sorted_predicated,k)

def mapk(actual, predicted, k=10):
    """
    Computes the mean average precision at k.
    This function computes the mean average prescision at k between two lists
    of lists of items.
    Parameters
    ----------
    actual : list
             A list of lists of elements that are to be predicted
             (order doesn't matter in the lists)
    predicted : list
                A list of lists of predicted elements
                (order matters in the lists)
    k : int, optional
        The maximum number of predicted elements
    Returns
    -------
    score : double
            The mean average precision at k over the input lists
    """
    return np.mean([apk(a,p,k) for a,p in zip(actual, predicted)])


def vmap(label, predicted_s):
    '''
    References
    ----------
    PASCAL VOC CHALLENGE (for binary classification problem)

    Parameters
    ----------
    label: A list of ground truth label (0 or 1)
    predicted_s: A list of predicted elements score
    Returns
    -------
    return mean average precision (for binary classification problem)
    '''
    if type(label) == list:
        label = np.array(label)
    if type(predicted_s) == list:
        predicted_s = np.array(predicted_s)

    if label.shape[0] == 0 or predicted_s.shape[0] == 0:
        return None

    # transform to array(int)
    if label.dtype != int:
        actual = label.astype(dtype=int)

    # absorb some error
    if np.max(label) > 1:
        label[np.where(label > 1)[0]] = 1
    if len(predicted_s.shape) > 1 and predicted_s.shape[1] > 1:
        predicted_s = predicted_s[:, 0]
    if len(predicted_s.shape) > 1:
        predicted_s = predicted_s.flatten()

    #data number
    num = label.size
    assert(label.size == predicted_s.size)

    #sort descend
    index = sorted(range(num), key=lambda k: predicted_s[k], reverse=True)

    label = label[index]
    predicted_s = predicted_s[index]

    positive_num = len(np.where(label == 1)[0])
    num_hits = 0.0

    #recall_thres = np.array(range(11)) * 0.1
    recall_thres = np.linspace(.0, 1.00, np.round((1.00 - .0) / .01) + 1, endpoint=True)
    rc = []
    pr = []
    for i in range(label.size):
        if predicted_s[i] <= -float("inf"):
            continue

        if label[i] == 1:
            num_hits += 1.0
            #recall
            recall = num_hits / positive_num

            rc.append(recall)
            pr.append(num_hits / (i + 1.0))

    for i in range(len(pr) - 1, 0, -1):
        if pr[i] > pr[i - 1]:
            pr[i - 1] = pr[i]

    inds = np.searchsorted(rc, recall_thres, side='left')
    q = [-1 for _ in range(len(inds))]
    try:
        for ri, pi in enumerate(inds):
            if pi == len(pr):
                q[ri] = pr[pi - 1]
            else:
                q[ri] = pr[pi]
    except:
        pass

    return float(np.mean(q))

#unit test
class TestAveragePrecision(unittest.TestCase):
    def test_apk(self):
        apk_val = apk(range(1, 6), [6, 4, 7, 1, 2], 2)
        self.assertAlmostEqual(apk_val,0.25)

        actual = [0,1,1,1,1,1,0,0]
        predicted_s = [0.1,0.5,0.4,0.3,0.7,0.2,0.8,0.6]

        self.assertAlmostEqual(s_apk(actual,predicted_s,2),0.25)
        self.assertAlmostEqual(map(actual, predicted_s), 0.270995670995671)

if __name__ == "__main__":
    unittest.main()


# -*- coding: UTF-8 -*-
# @Time    : 2020-06-25 23:29
# @File    : browser.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.ant.base import *
from antgo.ant.base import _pick_idle_port
# from antgo.dataflow import dataset
from antgo.dataflow.dataset.queue_dataset import *
from antgo.dataflow.recorder import *
from antvis.client.httprpc import *
from antgo.crowdsource.browser_server import *
import requests
import json
import traceback
import threading
import cv2


class BrowserDataRecorder(object):
  def __init__(self, rpc, maxsize=30):
    # self.queue = multiprocessing.Queue()              # 不设置队列最大缓冲
    self.prepare_queue = queue.Queue(maxsize=maxsize)
    self.dump_dir = ''
    self.dataset_flag = 'TRAIN'
    self.dataset_size = 0
    self.tag_dir = ''
    self.rpc = rpc

    # 5个线程，等待处理
    for _ in range(1):
      t = threading.Thread(target=self.asyn_record)
      t.daemon = True
      t.start()

  def _transfer_image(self, data):
    try:
      if len(data.shape) == 2:
        if data.dtype == np.uint8:
          transfer_result = data
        else:
          data_min = np.min(data)
          data_max = np.max(data)
          transfer_result = ((data - data_min) / (data_max - data_min) * 255).astype(np.uint8)
      else:
        assert (data.shape[2] == 3)
        transfer_result = data.astype(np.uint8)

      # save path
      if not os.path.exists(self.dump_dir):
        os.makedirs(self.dump_dir)

      file_name = '%s.png' % str(uuid.uuid4())
      image_path = os.path.join(self.dump_dir, file_name)
      cv2.imwrite(image_path, transfer_result)
      return file_name
    except:
      logger.error('Couldnt transfer image data.')
      return None

  def asyn_record(self):
    while True:
      val = self.prepare_queue.get()
      if val is None:
        break

      # 加入到队列中
      assert(type(val) == dict)

      # 重新组织数据格式
      data = {}
      id = None
      for key, value in val.items():
        if type(value) != dict:
          if key.lower() == 'id' or key.lower() == 'image_file':
            id = value
          continue

        data_name = key
        if data_name not in data:
          data[data_name] = {}

        if 'data' in value or 'DATA' in value:
          if 'data' in value:
            data[data_name]['data'] = value['data']
          else:
            data[data_name]['data'] = value['DATA']
        if 'type' in value or 'TYPE' in value:
          if 'type' in value:
            data[data_name]['type'] = value['type']
          else:
            data[data_name]['type'] = value['TYPE']
        
          if 'bboxes' in value and data[data_name]['type'] == 'IMAGE':
            data[data_name]['bboxes'] = value['bboxes']

        data[data_name]['title'] = data_name

      if id is None:
        logger.error('Missing id flag, please return {"id": ...}')
        continue

      for k, v in data.items():
        v['id'] = id

      # 转换数据到适合web
      web_data = []
      for name, body in data.items():
        if 'type' in body:
          if body['type'].upper() == 'IMAGE':
            body['width'] = body['data'].shape[1]
            body['height'] = body['data'].shape[0]
            body['data'] = '/static/data/%s'%self._transfer_image(body['data'])

        if 'type' not in body:
          body['type'] = 'STRING'

        if 'tag' not in body:
          body['tag'] = []

        body['dataset_flag'] = self.dataset_flag
        body['dataset_size'] = self.dataset_size
        web_data.append(body)

      self.rpc.browser.sample.fresh.post(samples=json.dumps([web_data]))

  def record(self, val):
    self.prepare_queue.put(val)

  def close(self):
    pass
  
class AntBrowser(AntBase):
  def __init__(self,
               ant_context,
               ant_name,
               ant_data_folder,
               ant_dump_dir,
               ant_token,
               ant_task_config,
               ant_dataset,
               **kwargs):
    super(AntBrowser, self).__init__(ant_name, ant_context, ant_token, **kwargs)
    self.ant_data_source = ant_data_folder
    self.ant_task_config = ant_task_config
    self.dataset_name = ant_dataset
    self.dump_dir = ant_dump_dir
    self.host_ip = self.context.params.system.get('ip', '127.0.0.1')
    self.host_port = self.context.params.system.get('port', 8901)
    self.rpc = None
    self._running_dataset = None
    self._running_task = None
    self.p = None

  @property
  def running_dataset(self):
    return self._running_dataset

  @property
  def running_task(self):
    return self._running_task

  def ping_until_ok(self):
    while True:
      content = self.rpc.ping.get()
      if content['status'] != 'ERROR':
          break
      # 暂停5秒钟，再进行尝试
      time.sleep(5)

  def wait_until_stop(self):
    self.p.join()

  def download(self):
    if not os.path.exists(os.path.join(self.dump_dir, 'check')):
      os.makedirs(os.path.join(self.dump_dir, 'check'))

    folder = os.path.join(self.dump_dir, 'check')
    file_name = f'check.json'
    self.rpc.browser.download(
      file_folder=folder,
      file_name=file_name
    )    
    with open(os.path.join(folder, file_name), 'r') as fp:
      content = json.load(fp)

      return content
  
  def waiting(self, until_exit=False):
    # 需要等待本轮标准完成
    while True:
      response = self.rpc.info.get()
      if response['status'] == 'ERROR':
        print('rpc error...')
        time.sleep(5)
        continue

      if response['content']['project_state']['stage'] == 'finish' and not until_exit:
        break
      # 等待10分钟后检查
      time.sleep(10)
  
  def start(self, *args, **kwargs):
    # 1.step 获得数据集解析
    running_ant_task = None
    if self.token is not None:
      # 1.1.step load challenge task
      response = mlogger.info.challenge.get(command=type(self).__name__)
      if response['status'] == 'ERROR':
        # invalid token
        logger.error('couldnt load challenge task')
        self.token = None
      elif response['status'] == 'SUSPEND':
        # prohibit submit challenge task frequently
        # submit only one in one week
        logger.error('prohibit submit challenge task frequently')
        exit(-1)
      elif response['status'] == 'OK':
        content = response['content']
        # maybe user token or task token
        if 'task' in content:
          challenge_task = create_task_from_json(content)
          if challenge_task is None:
            logger.error('couldnt load challenge task')
            exit(-1)
          running_ant_task = challenge_task
      else:
        # unknow error
        logger.error('unknow error')
        exit(-1)

    if running_ant_task is None:
        # 1.2.step load custom task
        if self.ant_task_config is not None:
            custom_task = create_task_from_xml(self.ant_task_config, self.context)
            if custom_task is None:
                logger.error('Couldnt load custom task.')
                exit(-1)
            running_ant_task = custom_task

    self._running_task = running_ant_task
    if running_ant_task is not None and running_ant_task.dataset is not None:
      self.dataset_name = running_ant_task.dataset_name
    dataset_cls = running_ant_task.dataset

    # 2.step 配置记录器
    self.rpc = HttpRpc("v1", "antgo/api", self.host_ip, self.host_port)
    self.context.recorder = \
      BrowserDataRecorder(rpc=self.rpc)

    # 3.step 启动浏览web服务
    browser_params = getattr(self.context.params, 'browser', None)
    tags = []
    if browser_params is not None:
      tags = browser_params.get('tags', [])

    # 3.1.step 准备web服务资源
    static_folder = '/'.join(os.path.dirname(__file__).split('/')[0:-1])
    if os.path.exists(self.dump_dir):
      shutil.rmtree(self.dump_dir)
    browser_static_dir = os.path.join(self.dump_dir, 'browser', 'static')
    shutil.copytree(os.path.join(static_folder, 'resource', 'app'), browser_static_dir)

    # 3.2.step 准备有效端口
    self.host_port = _pick_idle_port(self.host_port)

    # 3.3.step 状态
    train_offset, val_offset, test_offset = 0, 0, 0
    offset_configs = [{
      'dataset_flag': 'TRAIN',
      'dataset_offset': train_offset
    }, {
      'dataset_flag': 'VAL',
      'dataset_offset': val_offset
    }, {
      'dataset_flag': 'TEST',
      'dataset_offset': test_offset
    }]

    if self.context.params.browser is not None and \
      'offset' in self.context.params.browser.keys():
      if 'TRAIN' in self.context.params.browser.offset.keys():
        train_offset = self.context.params.browser.offset.TRAIN
        offset_config = {
          'dataset_flag': 'TRAIN',
          'dataset_offset': train_offset
        }
        offset_configs[0] = offset_config

    if self.context.params.browser is not None and \
      'offset' in self.context.params.browser.keys():
      if 'VAL' in self.context.params.browser.offset:
        val_offset = self.context.params.browser.offset.VAL
        offset_config = {
          'dataset_flag': 'VAL',
          'dataset_offset': val_offset
        }
        offset_configs[1] = offset_config

    if self.context.params.browser is not None and \
      'offset' in self.context.params.browser.keys():
      if 'TEST' in self.context.params.browser.offset:
        test_offset = self.context.params.browser.offset.TEST
        offset_config = {
          'dataset_flag': 'TEST',
          'dataset_offset': test_offset
        }
        offset_configs[2] = offset_config

    dataset_flag = 'train'
    if self.context.params.browser is not None and \
        'dataset_flag' in self.context.params.browser.keys():
      if self.context.params.browser.dataset_flag.lower() in ['train', 'val', 'test']:
        dataset_flag = \
          self.context.params.browser.dataset_flag.lower()

    data_json_file = kwargs.get('json_file', None)
    if dataset_cls is not None and data_json_file is None:
      train_dataset = \
        dataset_cls(dataset_flag, os.path.join(self.ant_data_source, self.dataset_name))
      self._running_dataset = train_dataset

    self.context.recorder.dataset_flag = dataset_flag.upper()
    self.context.recorder.dataset_size = \
      self.running_dataset.size if self.running_dataset is not None else self.context.params.browser.size
    self.context.recorder.dump_dir = os.path.join(self.dump_dir, 'browser', 'static', 'data')
    if not os.path.exists(self.context.recorder.dump_dir):
      os.makedirs(self.context.recorder.dump_dir)
    self.context.recorder.tag_dir = os.path.join(self.dump_dir, 'record')
    if not os.path.exists(self.context.recorder.tag_dir):
      os.makedirs(self.context.recorder.tag_dir)

    sample_meta = {}
    sample_list = []
    sample_folder = None
    if data_json_file is not None:
      # step1: 加载样本信息
      # 直接使用来自于data_json_file中的样本
      # 兼容两种常用的存储样本信息方式（1.纯json格式，所有样本以list形式存储；2.样本按行存储，每个样本的信息是json格式）
      try:
        # try 1 (纯json格式，所有样本以list形式存储)
        with open(data_json_file, 'r', encoding="utf-8") as fp:
          sample_list = json.load(fp)
          self.context.recorder.dataset_size = len(sample_list)
      except:
        # try 2 (样本信息按行存储，每个样本信息是json格式)
        with open(data_json_file, 'r', encoding="utf-8") as fp:
          sample_info_str = fp.readline()
          sample_info_str = sample_info_str.strip()
          
          while sample_info_str:
            sample_info = json.loads(sample_info_str)
            sample_list.append(sample_info)
            sample_info_str = fp.readline()
            sample_info_str = sample_info_str.strip()
            if sample_info_str == '':
              break        
            
      # step2: 尝试加载样本集meta信息
      sample_folder = os.path.dirname(data_json_file)
      data_meta_file = os.path.join(sample_folder, 'meta.json')
      if os.path.exists(data_meta_file):
        with open(data_meta_file, 'r', encoding="utf-8") as fp:
          sample_meta = json.load(fp)
              
    sample_offset = train_offset
    if dataset_flag == 'test':
      sample_offset = test_offset
    elif dataset_flag == 'val':
      sample_offset = val_offset
    logger.info('Browser %s dataset.' % dataset_flag)

    # 设置数据基本信息
    profile_config = {
      'dataset_flag': dataset_flag.upper(),
      'samples_num': self.context.recorder.dataset_size,
      'samples_num_checked': sample_offset
    }

    # 设置记录器偏移
    self.context.recorder.sample_index = sample_offset

    white_users = \
      self.context.params.browser.white_users.get() if self.context.params.browser.white_users is not None else None
    if len(white_users) == 0:
      white_users = None
    user_input = self.context.params.browser.user_input
    # 在独立进程中启动webserver
    self.p = \
      multiprocessing.Process(
        target=browser_server_start,
        args=(self.dump_dir,
              tags,
              self.host_port,
              offset_configs,
              profile_config,
              sample_folder, sample_list, sample_meta, user_input, white_users)
      )
    self.p.daemon = True
    self.p.start()

    # 等待直到http服务开启
    self.ping_until_ok()

    if self.context.is_interact_mode:
      return

    # 3.3.step 启动数据处理
    try:
      count = 0
      for data in self.context.data_processor.iterator(self.running_dataset):
        logger.info('Record data %d for browser.' % count)
        self.context.recorder.record(data)
        count += 1
    except Exception as e:
      traceback.print_exc()
      logger.info('Finish all records in browser %s dataset.' % dataset_flag)

    # 不结束webserver
    self.p.join()
    logger.info('Stop Browser.')


# -*- coding: UTF-8 -*-
# @Time    : 18-4-27
# @File    : demo_server.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

import tornado.httpserver
import tornado.ioloop
import tornado.options
from tornado.options import define, options
from tornado import web, gen
import tornado.web
from antgo.utils import logger
import os
import shutil
import json
import numpy as np
from antgo.utils.fs import *
from antgo.utils.encode import *
from PIL import Image
import uuid
import signal
from ..utils.serialize import loads,dumps
from antgo.crowdsource.utils import *
from tornado.concurrent import run_on_executor
import requests
from concurrent.futures import ThreadPoolExecutor
import functools
import base64
import threading
import time
from antgo.utils.utils import *
try:
    import queue
except:
    import Queue as queue


class BaseHandler(tornado.web.RequestHandler):
  executor = ThreadPoolExecutor(10)

  @property
  def response_dict(self):
    return self.settings.get('response_dict')

  @property
  def response_cond(self):
    return self.settings.get('response_cond')

  @property
  def demo_name(self):
    return self.settings.get('demo_name', '-')

  @property
  def demo_type(self):
    return self.settings.get('demo_type', '-')

  @property
  def demo_description(self):
    return self.settings.get('demo_description',{})

  @property
  def demo_dump(self):
    return self.settings.get('demo_dump', '')

  @property
  def demo_support_user_upload(self):
    return self.settings.get('support_user_upload', True)

  @property
  def demo_support_user_input(self):
    return self.settings.get('support_user_input', True)

  @property
  def demo_support_user_interaction(self):
    return self.settings.get('support_user_interaction', False)

  @property
  def demo_support_user_comment(self):
    return self.settings.get('support_user_comment', False)

  @property
  def html_template(self):
    return self.settings['html_template']
  
  @property
  def db(self):
    return self.settings['db']
  
  @property
  def port(self):
    return self.settings['port']

  @property
  def demo_constraint(self):
    constraint = self.settings.get('support_user_constraint', None)
    if constraint is None:
      constraint = {}

    return constraint

  @property
  def dataset_queue(self):
    return self.settings['dataset_queue']

  @property
  def request_waiting_time(self):
    return self.settings['request_waiting_time']

  def _transfer(self, data_key, data_value, data_type, demo_response, is_api_call):
    file_uuid = str(uuid.uuid4())
    if data_type == 'IMAGE':
      if os.path.exists(data_value):
        data = os.path.normpath(data_value)
        file_ext = data.split('/')[-1].split('.')[-1]
        if not is_api_call:
          shutil.copy(data, os.path.join(self.demo_dump, 'static', 'output', f'{file_uuid}.{file_ext}'))
          demo_response['DATA'].update(
            {data_key: {'DATA': '/static/output/%s' % f'{file_uuid}.{file_ext}', 'TYPE': 'IMAGE'}}
          )
        else:
          with open(data, 'rb') as fp:
            image_content = base64.b64encode(fp.read()).decode('utf-8')
            demo_response['DATA'].update(
              {data_key: {'DATA': image_content, 'TYPE': 'IMAGE'}}
            )
    elif data_type == 'VIDEO':
      if os.path.exists(data_value):
        data = os.path.normpath(data_value)
        file_ext = data.split('/')[-1].split('.')[-1]
        shutil.copy(data, os.path.join(self.demo_dump, 'static', 'output', f'{file_uuid}.{file_ext}'))
        demo_response['DATA'].update(
          {data_key: {'DATA': '/static/output/%s' % f'{file_uuid}.{file_ext}', 'TYPE': 'VIDEO'}}
        )
    elif data_type == 'AUDIO':
      if os.path.exists(data_value):
        data = os.path.normpath(data_value)
        file_ext = data.split('/')[-1].split('.')[-1]
        shutil.copy(data, os.path.join(self.demo_dump, 'static', 'output', f'{file_uuid}.{file_ext}'))
        demo_response['DATA'].update(
          {data_key: {'DATA': '/static/output/%s' % f'{file_uuid}.{file_ext}', 'TYPE': 'AUDIO'}}
        )
    elif data_type == 'FILE':
      if os.path.exists(data_value):
        data = os.path.normpath(data_value)
        file_ext = data.split('/')[-1].split('.')[-1]
        shutil.copy(data, os.path.join(self.demo_dump, 'static', 'output', f'{file_uuid}.{file_ext}'))
        demo_response['DATA'].update(
          {data_key: {'DATA': '/static/output/%s' % f'{file_uuid}.{file_ext}', 'TYPE': 'FILE'}}
        )
    else:
      # string
      demo_response['DATA'].update({data_key: {'DATA': str(data_value), 'TYPE': 'STRING'}})

    return demo_response

  @run_on_executor
  def waitingResponse(self, data_id, time_stamp, is_api_call=False):
    self.response_cond.acquire()
    data_response = None
    while True:
      for k, v in self.response_dict.items():
        if k == data_id:
          data_response = v
          break

      if data_response is not None:
        break
      else:
        current_time = time.time()
        if current_time > time_stamp + self.request_waiting_time:
          break

        if not self.response_cond.wait(30):
          # 等待10s，还无法获得锁，返回无效数据
          break

    # 执行到这里，存在两种情况，（1）已经获得结果；（2）已经超时，但没有获得结果
    if data_id in self.response_dict:
      self.response_dict.pop(data_id)
    self.response_cond.release()

    if data_response is None:
      return {}

    processed_data = data_response['data']
    if len(processed_data) == 0:
      return {}

    # build output folder (static/output)
    if not os.path.exists(os.path.join(self.demo_dump, 'static', 'output')):
      os.makedirs(os.path.join(self.demo_dump, 'static', 'output'))

    demo_response = {'DATA': {'RESULT': '', 'TYPE': 'STRING'}}
    for data in processed_data[0]:
      item_type = data['type']
      item_data = data['data']
      item_title = data['title']

      demo_response = self._transfer(item_title, item_data, item_type, demo_response, is_api_call)
    return demo_response

  @run_on_executor
  def asynProcess(self, preprocess_type, data):
    if preprocess_type == 'DOWNLOAD':
      try:
        # 1.step 下载(如果有必要)
        pic = requests.get(data['url'], timeout=7)
        download_path = os.path.join(self.demo_dump, 'static', 'input', data['file_name'])
        with open(download_path, 'wb') as fp:
          fp.write(pic.content)

        # 2.step 检查
        # 2.1.step 检查文件大小
        if 'file_size' in self.demo_constraint:
          max_file_size = self.demo_constraint['file_size']
          fsize = os.path.getsize(download_path) / float(1024 * 1024)
          if round(fsize,2) > max_file_size:
            return {'status': 400,
                      'code': 'InvalidImageSize', 
                      'message': 'The input file size is too large (>%f MB)'%float(max_file_size)}

        # 2.2.step 检查文件格式
        # 图片格式检测（图片文件后缀可能不对）
        download_path = os.path.normpath(download_path)
        file_type = imghdr.what(download_path)
        if file_type in ['jpeg', 'png', 'gif', 'bmp']:
          # 图像文件
          file_name = '%s.%s'%(data['file_name'], file_type)
          os.rename(download_path, os.path.join('/'.join(download_path.split('/')[0:-1]), file_name))
          download_path = os.path.join('/'.join(download_path.split('/')[0:-1]), file_name)
        else:
          # 非图像文件
          file_type = download_path.split('.')[-1]

        if 'file_type' in self.demo_constraint:
          if file_type not in self.demo_constraint['file_type']:
            return {'status': 400,
                      'code': 'InvalidImageFormat', 
                      'message': 'The input file is not in a valid image format that the service can support'}

        return {
          'status': 200,
          'path': download_path
        }
      except:
        print('Fail to download %s'%data['url'])
        return {
          'status': 500,
          'code': 'UnkownError'
        }   
    elif preprocess_type == 'RECEIVE':
      try:
        # 1.step 保存文件
        file_path = data['file_path']
        file_data = data['file_data']
        with open(file_path, 'wb') as fp:
          fp.write(file_data)
        
        # 2.step 检查
        # 2.1.step 检查文件大小
        if 'file_size' in self.demo_constraint:
          max_file_size = self.demo_constraint['file_size']
          fsize = os.path.getsize(file_path) / float(1024 * 1024)
          if round(fsize, 2) > max_file_size:
            return {'status': 400,
                      'code': 'InvalidImageSize', 
                      'message': 'The input file size is too large (>%f MB)'%float(max_file_size)}

        # 2.2.step 检查文件格式
        # 图片格式检测（图片文件后缀可能不对）
        file_path = os.path.normpath(file_path)
        file_type = imghdr.what(file_path)
        if file_type in ['jpeg', 'png', 'gif', 'bmp']:
          # 图像文件
          file_name = '%s.%s'%(file_path.split('/')[-1].split('.')[0], file_type)
          os.rename(file_path, os.path.join('/'.join(file_path.split('/')[0:-1]), file_name))
          file_path = os.path.join('/'.join(file_path.split('/')[0:-1]), file_name)
        else:
          # 非图像文件，通过获取扩展名作为文件类型
          file_type = file_path.split('.')[-1]

        if 'file_type' in self.demo_constraint:
          if file_type not in self.demo_constraint['file_type']:
            return {'status': 400,
                      'code': 'InvalidImageFormat', 
                      'message': 'The input file is not in a valid image format that the service can support'}
        return {
          'status': 200,
          'path': file_path
        }
      except:
        return {
          'status': 500,
          'code': 'UnkownError'
        }   
    elif preprocess_type == 'PACKAGE':
        data_path = os.path.normpath(data['path'])
        data_name = data_path.split('/')[-1]
        ext_name = data_path.split('/')[-1].split('.')[-1].lower()

        # 根据后缀判断上传的数据类型
        if ext_name in ['jpg', 'jpeg', 'png', 'bmp', 'gif']:
          return {'data': (data_path, data_name, 'IMAGE'), 'status': 200}
        elif ext_name in ['mp4', 'avi', 'mov']:
          return {'data': (data_path, data_name, 'VIDEO'), 'status': 200}
        elif ext_name in ['txt']:
          return {'data': (data_path, data_name, 'FILE'), 'status': 200}
        else:
          # TODO: support video and sound
          logger.warn('dont support file type %s' % ext_name)
          return {'status': 400, 'code': 'InvalidPackage', 'message': 'Fail package'}
    elif preprocess_type == 'API_QUERY':
      # 1.step base64解码
      # format {'image': '', 'video': None, 'params': [{'data': ,'type': , 'name': ,},{}]}
      data_list = data
      if type(data_list) == dict:
        data_list = [data_list]

      multi_images = []
      for data in data_list:
        image_str = None
        if 'image' in data:
          image_str = data['image']

        if image_str is None or image_str == '':
          return {
            'status': 400,
            'code': 'InvalidData',
            'message': 'Missing query data'}

        image_b = base64.b64decode(image_str)
        multi_images.append(image_b)

      if len(multi_images) == 1:
        multi_images = multi_images[0]
      params = data.get('params', {})
      return {
        'status': 200,
        'data': {'image': multi_images, 'params': params}
      }


class IndexHandler(BaseHandler):
  def get(self):
    image_history_data = []

    history_folder = os.path.join(self.demo_dump, 'static', 'input')
    if os.path.exists(history_folder):
      for f in os.listdir(history_folder):
        ext_name = f.split('.')[-1]
        if ext_name.lower() in ['jpg', 'jpeg', 'png', 'gif']:
          image_history_data.append('/static/input/%s'%f)
    
    input_filter = ''

    if 'file_type' in self.demo_constraint:
      for support_format in self.demo_constraint['file_type']:
        if support_format.lower() in ['jpg', 'jpeg', 'png', 'gif']:
          input_filter += 'image/%s,'%support_format.lower()
        elif support_format.lower() in ['mp4','mov','avi']:
          input_filter += 'video/%s,'%support_format.lower()
        elif support_format.lower() in ['mp3', 'wav']:
          input_filter += 'audio/%s,'%support_format.lower()
        else:
          input_filter += '%s,'%support_format.lower()

    if len(input_filter) > 0:
      input_filter = input_filter[0:-1]
    else:
      input_filter = '*'

    self.render(self.html_template, demo={'name': self.demo_name,
                                          'type': self.demo_type,
                                          'description': self.demo_description,
                                          'upload': self.demo_support_user_upload,
                                          'upload_accept': input_filter,
                                          'input': self.demo_support_user_input,
                                          'image_history': image_history_data})


class ClientQueryHandler(BaseHandler):
  @gen.coroutine
  def post(self):
    # 0.step check support status
    if not self.demo_support_user_upload:
      self.set_status(500)
      self.write(json.dumps({'code': 'InvalidSupport', 'message': 'demo dont support upload'}))
      self.finish()
      return

    upload_path = os.path.join(self.demo_dump, 'static', 'input')
    if not os.path.exists(upload_path):
      os.makedirs(upload_path)

    # 1.step parse query data
    model_datas = []
    model_names = []
    model_data_types = []
    request_param = None

    data = self.get_argument('DATA', None)
    data_type = self.get_argument('DATA_TYPE', None)
    if data is not None and data_type is not None:
      if data_type == 'URL':
        # 下载
        file_name = str(uuid.uuid4())
        result = yield self.asynProcess('DOWNLOAD', {'url': data, 'file_name': file_name})
        if result['status'] != 200:
          self.set_status(result['status'])
          self.write(json.dumps(result))
          return

        # 获得下载后的路径
        file_path = result['path']

        # 打包
        result = yield self.asynProcess('PACKAGE', {'path': file_path})
        if result['status'] != 200:
          self.set_status(result['status'])
          self.write(json.dumps(result))
          return
        
        model_data, model_name, model_data_type = result['data']

        # 保存
        model_datas.append(model_data)
        model_names.append(model_name)
        model_data_types.append(model_data_type)
    else:
      if len(self.request.files) == 0:
        self.set_status(400)
        self.write(json.dumps({'code': 'InvalidUploadFile', 
                                'message': 'The input file is not uploaded correctly'}))
        self.finish()
        return

      file_paths = []
      file_names = []
      for _, meta in self.request.files.items():
        _file_name = '%s_%s' % (str(uuid.uuid4()), meta[0]['filename'])
        _file_path = os.path.join(upload_path, _file_name)
        
        # 接受文件
        result = yield self.asynProcess('RECEIVE', {'file_path': _file_path, 'file_data': meta[0]['body']})
        if result['status'] != 200:
          self.set_status(result['status'])
          result['code'] = 'InvalidUpload'
          result['message'] = 'Couldnt upload request file'
          self.write(json.dumps(result))
          return

        _file_path = result['path']
        _file_name = _file_path.split('/')

        file_paths.append(_file_path)
        file_names.append(_file_name)

        # 仅支持一个文件处理
        break

      # 打包
      result = yield self.asynProcess('PACKAGE', {'path': file_paths[0]})
      if result['status'] != 200:
        self.set_status(result['status'])
        self.write(json.dumps(result))
        return
      
      model_data, model_name, model_data_type = result['data']
      model_datas.append(model_data)
      model_names.append(model_name)
      model_data_types.append(model_data_type)

      request_param = self.get_argument('params', None)
      if request_param is not None:
        request_param = json.loads(request_param)

    if request_param is None:
      request_param = {}

    request_param.update({
      'id': model_names[0]
    })
    
    # push to backend
    model_input = (model_datas[0], model_data_types[0], request_param)

    # 发送到处理队列
    self.dataset_queue.put(model_input)
    
    # 异步等待响应(设置等待队列，并在异步线程中等待)
    data_response = yield self.waitingResponse(model_names[0], time.time())

    # 绑定输入数据,返回
    data_response['INPUT_TYPE'] = model_data_types[0]
    if model_data_types[0] in ['FILE', 'IMAGE', 'AUDIO', 'VIDEO']:
      data_response['INPUT'] = f'/static/input/{model_names[0]}'
    else:
      data_response['INPUT'] = model_datas[0]

    # 返回成功
    self.write(json.dumps(data_response))
    self.finish()


class ClientCliQueryHandler(BaseHandler):
  @gen.coroutine
  def post(self):
    request_data = self.get_argument('data', None)
    if request_data is None:
      self.set_status(400)
      self.write(json.dumps({'code': 'InvalidQuery', 'message': 'Invalid query data'}))
      return

    # 解析请求字符串
    request_data = json.loads(request_data)

    '''
    format: {'data': {'image': '', 'params': [{'data': ,'type': , 'name': ,},{}]}, 'time': ,}
    '''
    # query_time = request_data['time']
    query_data = request_data['data']

    result = yield self.asynProcess('API_QUERY', query_data)
    if result['status'] != 200:
        self.set_status(result['status'])
        self.write(json.dumps(result))
        return

    data_id = str(uuid.uuid4())
    request_param = {'id': data_id}
    image = result['data']['image']
    request_param.update(result['data']['params'])

    # push to backend
    data_type = 'IMAGE_MEMORY'
    if type(image) == list:
      data_type = 'IMAGE_MULTI_MEMORY'
    model_input = (image, data_type, request_param)

    # 发送到处理队列
    self.dataset_queue.put(model_input)
    
    # 异步等待响应(设置等待队列，并在异步线程中等待)
    data_response = yield self.waitingResponse(data_id, time.time(), True)

    # 返回成功
    self.write(json.dumps(data_response))
    self.finish()


class ClientDownloadHandler(BaseHandler):
  @gen.coroutine
  def get(self):
    file_name = self.get_argument('filename', None)
    if file_name is None:
      return

    download_path = os.path.join(self.settings.get('static_path'), 'output', file_name)
    if not os.path.exists(download_path):
      return

    # Content-Type
    self.set_header('Content-Type', 'application/octet-stream')
    self.set_header('Content-Disposition', 'attachment; filename=' + file_name)

    buffer_size = 64 * 1024
    with open(download_path, 'rb') as fp:
      content = fp.read()
      content_size = len(content)
      buffer_segments = content_size // buffer_size
      for buffer_seg_i in range(buffer_segments):
        buffer_data = content[buffer_seg_i * buffer_size: (buffer_seg_i + 1) * buffer_size]
        yield self.write(buffer_data)

      yield self.write(content[buffer_segments * buffer_size:])

    self.finish()


class ClientResponseHandler(BaseHandler):
  @gen.coroutine
  def post(self):
    # 返回数据
    # data type: 'FILE', 'STRING', 'IMAGE', 'VIDEO', 'AUDIO'
    # data:       PATH,  '',        PATH,    PATH,    PATH
    response_data = self.get_argument('response', None)
    if response_data is None:
      return
    
    response_data = json.loads(response_data)

    # 添加进入结果池
    self.response_cond.acquire()
    self.response_dict[response_data['id']] = response_data
    self.response_cond.notify_all()
    self.response_cond.release()


class ClientFileUploadAndProcessHandler(BaseHandler):
  @gen.coroutine
  def post(self):
    # 0.step check support status
    if not self.demo_support_user_upload:
      self.set_status(500)
      self.write(json.dumps({'code': 'InvalidSupport', 
                            'message': 'demo dont support upload'}))
      self.finish()
      return

    # 1.step receive client upload file
    upload_path = os.path.join(self.demo_dump, 'static', 'input')
    if not os.path.exists(upload_path):
      os.makedirs(upload_path)

    file_metas = self.request.files.get('file', None)
    if not file_metas:
      self.set_status(400)
      self.write(json.dumps({'code': 'InvalidUploadFile', 
                              'message': 'The input file is not uploaded correctly'}))
      self.finish()
      return

    file_paths = []
    file_names = []
    for meta in file_metas:
      # 对上传的文件创建唯一标识名字
      _file_name = '%s_%s'%(str(uuid.uuid4()), meta['filename'])
      _file_path = os.path.join(upload_path, _file_name)

      # 接受文件
      result = yield self.asynProcess('RECEIVE', {
                  'file_path': _file_path,
                  'file_data': meta['body']
                })
      if result['status'] != 200:
        self.set_status(result['status'])
        result['code'] = 'InvalidUpload'
        result['message'] = 'Couldnt upload request file'
        self.write(json.dumps(result))
        return

      _file_path = result['path']
      _file_name = _file_path.split('/')
      file_paths.append(_file_path)
      file_names.append(_file_name)

      # 仅支持单文件上传，处理
      break

    model_datas = []
    model_data_names = []
    model_data_types = []
    # 打包
    result = yield self.asynProcess('PACKAGE', {'path': file_paths[0]})
    if result['status'] != 200:
      self.set_status(result['status'])
      self.write(json.dumps(result))
      return
    
    model_data, model_data_name, model_data_type = result['data']
    model_datas.append(model_data)
    model_data_names.append(model_data_name)
    model_data_types.append(model_data_type)

    request_param = {'id': model_data_names[0]}
    model_input = (model_datas[0], model_data_types[0], request_param)

    # 发送到处理队列
    self.dataset_queue.put(model_input)
    
    # 异步等待响应(设置等待队列，并在异步线程中等待)
    data_response = yield self.waitingResponse(model_data_names[0], time.time())

    # 绑定输入数据,返回
    data_response['INPUT_TYPE'] = model_data_types[0]
    if model_data_types[0] in ['FILE', 'IMAGE', 'AUDIO', 'VIDEO']:
      data_response['INPUT'] = '/static/input/%s' % model_data_names[0]
    else:
      data_response['INPUT'] = model_datas[0]

    # 返回成功
    self.write(json.dumps(data_response))
    self.finish()


class ClientCommentHandler(BaseHandler):
  def post(self):
    self.set_status(201)
    self.finish()


class PrefixRedirectHandler(BaseHandler):
  def get(self):
    static_pi = self.request.uri.find('static')
    path = self.request.uri[static_pi:]
    self.redirect('http://127.0.0.1:%d/%s'%(self.port, path), permanent=False)


class PingApiHandler(BaseHandler):
  @gen.coroutine
  def get(self, *args, **kwargs):
    self.set_status(status_code)

    if status is None:
      status = 'OK'
      if status_code not in [200, 201, 204]:
        status = 'ERROR'

    self.write(json.dumps({
      'status': status,
      'message': message,
      'content': content
    }))


class GracefulExitException(Exception):
  @staticmethod
  def sigterm_handler(signum, frame):
    raise GracefulExitException()


def demo_server_start(demo_name,
                      demo_type,
                      demo_config,
                      dataset_queue,
                      request_waiting_time=30):
  # register sig
  signal.signal(signal.SIGTERM, GracefulExitException.sigterm_handler)
  
  try:
    # 0.step define http server port
    define('port', default=demo_config['port'], help='run on port')

    # 1.step prepare static resource
    static_folder = '/'.join(os.path.dirname(__file__).split('/')[0:-1])
    demo_static_dir = os.path.join(demo_config['dump_dir'], 'static')
    if not os.path.exists(demo_static_dir):
      os.makedirs(demo_static_dir)
  
    for static_file in os.listdir(os.path.join(static_folder, 'resource', 'static')):
      if static_file[0] == '.':
        continue
    
      shutil.copy(os.path.join(static_folder, 'resource', 'static', static_file), demo_static_dir)
  
    # 2.step prepare html template
    demo_tempate_dir = os.path.join(demo_config['dump_dir'], 'templates')
  
    if not os.path.exists(demo_tempate_dir):
      os.makedirs(demo_tempate_dir)

    html_template = demo_config['html_template']
    if html_template is None or html_template == '':
      html_template = 'demo.html'
  
      if not os.path.exists(os.path.join(demo_tempate_dir, html_template)):
        assert(os.path.exists(os.path.join(static_folder, 'resource', 'templates',html_template)))
        shutil.copy(os.path.join(static_folder, 'resource', 'templates',html_template), demo_tempate_dir)
    else:
      shutil.copy(html_template, demo_tempate_dir)

    tornado.options.parse_command_line()
    response_dict = {}
    settings = {
      'template_path': demo_tempate_dir,
      'static_path': demo_static_dir,
      'html_template': html_template,
      'port': demo_config['port'],
      'demo_dump': demo_config['dump_dir'],
      'demo_name': demo_name,
      'demo_type': demo_type,
      'demo_description': demo_config['description_config'],
      'support_user_upload': demo_config['interaction']['support_user_upload'],
      'support_user_input': demo_config['interaction']['support_user_input'],
      'support_user_interaction': demo_config['interaction']['support_user_interaction'],
      'support_user_constraint': demo_config['interaction']['support_user_constraint'],
      'dataset_queue': dataset_queue,
      'request_waiting_time': request_waiting_time,
      'response_dict': response_dict,
      'response_cond': threading.Condition()
    }
    app = tornado.web.Application(handlers=[
      (r"/", IndexHandler),
      (r"/antgo/api/demo/query/", ClientQueryHandler),
      (r"/antgo/api/demo/comment/", ClientCommentHandler),
      (r"/antgo/api/demo/response/", ClientResponseHandler),
      (r"/antgo/api/demo/cli_query/", ClientCliQueryHandler),
      (r"/antgo/api/demo/download/", ClientDownloadHandler),
      (r"/antgo/api/demo/submit/", ClientFileUploadAndProcessHandler),
      (r"/antgo/api/ping/", PingApiHandler),
      (r"/.*/static/.*", PrefixRedirectHandler)
    ], **settings)
    http_server = tornado.httpserver.HTTPServer(app)
    http_server.listen(options.port)
  
    logger.info('demo is providing server on port %d'%demo_config['port'])
    tornado.ioloop.IOLoop.instance().start()
    logger.info('demo stop server')
  except GracefulExitException:
    logger.info('demo server exit')
    sys.exit(0)




import torch
import torch.nn as nn

from antgo.framework.helper.models.builder import LOSSES
from .utils import weight_reduce_loss


def dice_loss(pred,
              target,
              weight=None,
              eps=1e-3,
              reduction='mean',
              avg_factor=None):
    """Calculate dice loss, which is proposed in
    `V-Net: Fully Convolutional Neural Networks for Volumetric
    Medical Image Segmentation <https://arxiv.org/abs/1606.04797>`_.

    Args:
        pred (torch.Tensor): The prediction, has a shape (n, *)
        target (torch.Tensor): The learning label of the prediction,
            shape (n, *), same shape of pred.
        weight (torch.Tensor, optional): The weight of loss for each
            prediction, has a shape (n,). Defaults to None.
        eps (float): Avoid dividing by zero. Default: 1e-3.
        reduction (str, optional): The method used to reduce the loss into
            a scalar. Defaults to 'mean'.
            Options are "none", "mean" and "sum".
        avg_factor (int, optional): Average factor that is used to average
            the loss. Defaults to None.
    """

    input = pred.flatten(1)
    target = target.flatten(1).float()

    a = torch.sum(input * target, 1)
    b = torch.sum(input * input, 1) + eps
    c = torch.sum(target * target, 1) + eps
    d = (2 * a) / (b + c)
    loss = 1 - d
    if weight is not None:
        assert weight.ndim == loss.ndim
        assert len(weight) == len(pred)
    loss = weight_reduce_loss(loss, weight, reduction, avg_factor)
    return loss


@LOSSES.register_module()
class DiceLoss(nn.Module):

    def __init__(self,
                 use_sigmoid=True,
                 activate=True,
                 reduction='mean',
                 loss_weight=1.0,
                 eps=1e-3):
        """`Dice Loss, which is proposed in
        `V-Net: Fully Convolutional Neural Networks for Volumetric
         Medical Image Segmentation <https://arxiv.org/abs/1606.04797>`_.

        Args:
            use_sigmoid (bool, optional): Whether to the prediction is
                used for sigmoid or softmax. Defaults to True.
            activate (bool): Whether to activate the predictions inside,
                this will disable the inside sigmoid operation.
                Defaults to True.
            reduction (str, optional): The method used
                to reduce the loss. Options are "none",
                "mean" and "sum". Defaults to 'mean'.
            loss_weight (float, optional): Weight of loss. Defaults to 1.0.
            eps (float): Avoid dividing by zero. Defaults to 1e-3.
        """

        super(DiceLoss, self).__init__()
        self.use_sigmoid = use_sigmoid
        self.reduction = reduction
        self.loss_weight = loss_weight
        self.eps = eps
        self.activate = activate

    def forward(self,
                pred,
                target,
                weight=None,
                reduction_override=None,
                avg_factor=None):
        """Forward function.

        Args:
            pred (torch.Tensor): The prediction, has a shape (n, *).
            target (torch.Tensor): The label of the prediction,
                shape (n, *), same shape of pred.
            weight (torch.Tensor, optional): The weight of loss for each
                prediction, has a shape (n,). Defaults to None.
            avg_factor (int, optional): Average factor that is used to average
                the loss. Defaults to None.
            reduction_override (str, optional): The reduction method used to
                override the original reduction method of the loss.
                Options are "none", "mean" and "sum".

        Returns:
            torch.Tensor: The calculated loss
        """

        assert reduction_override in (None, 'none', 'mean', 'sum')
        reduction = (
            reduction_override if reduction_override else self.reduction)

        if self.activate:
            if self.use_sigmoid:
                pred = pred.sigmoid()
            else:
                raise NotImplementedError

        loss = self.loss_weight * dice_loss(
            pred,
            target,
            weight,
            eps=self.eps,
            reduction=reduction,
            avg_factor=avg_factor)

        return loss



# -*- coding: UTF-8 -*-
# @Time    : 2022/9/24 20:39
# @File    : ocr.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import copy
import os
import warnings
from argparse import ArgumentParser, Namespace
from pathlib import Path

import mmcv
import numpy as np
import torch
from mmcv.image.misc import tensor2imgs
from mmcv.runner import load_checkpoint
from mmcv.utils.config import Config
from PIL import Image
import sys
import subprocess

try:
    import tesserocr
except ImportError:
    tesserocr = None

from mmocr.apis import init_detector
from mmocr.apis.inference import model_inference
from mmocr.core.visualize import det_recog_show_result
from mmocr.datasets.kie_dataset import KIEDataset
from mmocr.datasets.pipelines.crop import crop_img
from mmocr.models import build_detector
from mmocr.models.textdet.detectors import TextDetectorMixin
from mmocr.models.textrecog.recognizer import BaseRecognizer
from mmocr.utils import is_type_list
from mmocr.utils.box_util import stitch_boxes_into_lines
from mmocr.utils.fileio import list_from_file
from mmocr.utils.model import revert_sync_batchnorm


class MMOCR:
    def __init__(self,
                 det_config='',
                 det_ckpt='',
                 recog_config='',
                 recog_ckpt='',
                 config_dir=os.path.join(str(Path.cwd()), 'configs/'),
                 device=None,
                 **kwargs):

        self.device = device
        if self.device is None:
            self.device = torch.device(
                'cuda' if torch.cuda.is_available() else 'cpu')

        self.detect_model = init_detector(
            det_config, det_ckpt, device=self.device)
        self.detect_model = revert_sync_batchnorm(self.detect_model)

        self.recog_model = init_detector(
            recog_config, recog_ckpt, device=self.device)
        self.recog_model = revert_sync_batchnorm(self.recog_model)

        self.kie_model = None
        # Attribute check
        for model in list(filter(None, [self.recog_model, self.detect_model])):
            if hasattr(model, 'module'):
                model = model.module


    def readtext(self,
                 img,
                 output=None,
                 details=False,
                 export=None,
                 export_format='json',
                 batch_mode=False,
                 recog_batch_size=0,
                 det_batch_size=0,
                 single_batch_size=0,
                 imshow=False,
                 print_result=False,
                 merge=False,
                 merge_xdist=20,
                 **kwargs):
        args = locals().copy()
        [args.pop(x, None) for x in ['kwargs', 'self']]
        args = Namespace(**args)

        # Input and output arguments processing
        self._args_processing(args)
        self.args = args

        pp_result = None

        # Send args and models to the MMOCR model inference API
        # and call post-processing functions for the output
        if self.detect_model and self.recog_model:
            det_recog_result = self.det_recog_kie_inference(
                self.detect_model, self.recog_model, kie_model=self.kie_model)
            pp_result = self.det_recog_pp(det_recog_result)
        else:
            for model in list(
                    filter(None, [self.recog_model, self.detect_model])):
                result = self.single_inference(model, args.arrays,
                                               args.batch_mode,
                                               args.single_batch_size)
                pp_result = self.single_pp(result, model)

        return pp_result

    # Post processing function for end2end ocr
    def det_recog_pp(self, result):
        final_results = []
        args = self.args
        for arr, output, export, det_recog_result in zip(
                args.arrays, args.output, args.export, result):
            if output or args.imshow:
                if self.kie_model:
                    res_img = det_recog_show_result(arr, det_recog_result)
                else:
                    res_img = det_recog_show_result(
                        arr, det_recog_result, out_file=output)
                if args.imshow and not self.kie_model:
                    mmcv.imshow(res_img, 'inference results')
            if not args.details:
                simple_res = {}
                simple_res['filename'] = det_recog_result['filename']
                simple_res['text'] = [
                    x['text'] for x in det_recog_result['result']
                ]
                final_result = simple_res
            else:
                final_result = det_recog_result
            if export:
                mmcv.dump(final_result, export, indent=4)
            if args.print_result:
                print(final_result, end='\n\n')
            final_results.append(final_result)
        return final_results

    # Post processing function for separate det/recog inference
    def single_pp(self, result, model):
        for arr, output, export, res in zip(self.args.arrays, self.args.output,
                                            self.args.export, result):
            if export:
                mmcv.dump(res, export, indent=4)
            if output or self.args.imshow:
                if model == 'Tesseract_det':
                    res_img = TextDetectorMixin(show_score=False).show_result(
                        arr, res, out_file=output)
                elif model == 'Tesseract_recog':
                    res_img = BaseRecognizer.show_result(
                        arr, res, out_file=output)
                else:
                    res_img = model.show_result(arr, res, out_file=output)
                if self.args.imshow:
                    mmcv.imshow(res_img, 'inference results')
            if self.args.print_result:
                print(res, end='\n\n')
        return result

    def generate_kie_labels(self, result, boxes, class_list):
        idx_to_cls = {}
        if class_list is not None:
            for line in list_from_file(class_list):
                class_idx, class_label = line.strip().split()
                idx_to_cls[class_idx] = class_label

        max_value, max_idx = torch.max(result['nodes'].detach().cpu(), -1)
        node_pred_label = max_idx.numpy().tolist()
        node_pred_score = max_value.numpy().tolist()
        labels = []
        for i in range(len(boxes)):
            pred_label = str(node_pred_label[i])
            if pred_label in idx_to_cls:
                pred_label = idx_to_cls[pred_label]
            pred_score = node_pred_score[i]
            labels.append((pred_label, pred_score))
        return labels

    def visualize_kie_output(self,
                             model,
                             data,
                             result,
                             out_file=None,
                             show=False):
        """Visualizes KIE output."""
        img_tensor = data['img'].data
        img_meta = data['img_metas'].data
        gt_bboxes = data['gt_bboxes'].data.numpy().tolist()
        if img_tensor.dtype == torch.uint8:
            # The img tensor is the raw input not being normalized
            # (For SDMGR non-visual)
            img = img_tensor.cpu().numpy().transpose(1, 2, 0)
        else:
            img = tensor2imgs(
                img_tensor.unsqueeze(0), **img_meta.get('img_norm_cfg', {}))[0]
        h, w, _ = img_meta.get('img_shape', img.shape)
        img_show = img[:h, :w, :]
        model.show_result(
            img_show, result, gt_bboxes, show=show, out_file=out_file)

    # End2end ocr inference pipeline
    def det_recog_kie_inference(self, det_model, recog_model, kie_model=None):
        end2end_res = []
        # Find bounding boxes in the images (text detection)
        det_result = self.single_inference(det_model, self.args.arrays,
                                           self.args.batch_mode,
                                           self.args.det_batch_size)
        bboxes_list = [res['boundary_result'] for res in det_result]

        if kie_model:
            kie_dataset = KIEDataset(
                dict_file=kie_model.cfg.data.test.dict_file)

        # For each bounding box, the image is cropped and
        # sent to the recognition model either one by one
        # or all together depending on the batch_mode
        for filename, arr, bboxes, out_file in zip(self.args.filenames,
                                                   self.args.arrays,
                                                   bboxes_list,
                                                   self.args.output):
            img_e2e_res = {}
            img_e2e_res['filename'] = filename
            img_e2e_res['result'] = []
            box_imgs = []
            for bbox in bboxes:
                box_res = {}
                box_res['box'] = [round(x) for x in bbox[:-1]]
                box_res['box_score'] = float(bbox[-1])
                box = bbox[:8]
                if len(bbox) > 9:
                    min_x = min(bbox[0:-1:2])
                    min_y = min(bbox[1:-1:2])
                    max_x = max(bbox[0:-1:2])
                    max_y = max(bbox[1:-1:2])
                    box = [
                        min_x, min_y, max_x, min_y, max_x, max_y, min_x, max_y
                    ]
                box_img = crop_img(arr, box)
                if self.args.batch_mode:
                    box_imgs.append(box_img)
                else:
                    if recog_model == 'Tesseract_recog':
                        recog_result = self.single_inference(
                            recog_model, box_img, batch_mode=True)
                    else:
                        recog_result = model_inference(recog_model, box_img)
                    text = recog_result['text']
                    text_score = recog_result['score']
                    if isinstance(text_score, list):
                        text_score = sum(text_score) / max(1, len(text))
                    box_res['text'] = text
                    box_res['text_score'] = text_score
                img_e2e_res['result'].append(box_res)

            if self.args.batch_mode:
                recog_results = self.single_inference(
                    recog_model, box_imgs, True, self.args.recog_batch_size)
                for i, recog_result in enumerate(recog_results):
                    text = recog_result['text']
                    text_score = recog_result['score']
                    if isinstance(text_score, (list, tuple)):
                        text_score = sum(text_score) / max(1, len(text))
                    img_e2e_res['result'][i]['text'] = text
                    img_e2e_res['result'][i]['text_score'] = text_score

            if self.args.merge:
                img_e2e_res['result'] = stitch_boxes_into_lines(
                    img_e2e_res['result'], self.args.merge_xdist, 0.5)

            if kie_model:
                annotations = copy.deepcopy(img_e2e_res['result'])
                # Customized for kie_dataset, which
                # assumes that boxes are represented by only 4 points
                for i, ann in enumerate(annotations):
                    min_x = min(ann['box'][::2])
                    min_y = min(ann['box'][1::2])
                    max_x = max(ann['box'][::2])
                    max_y = max(ann['box'][1::2])
                    annotations[i]['box'] = [
                        min_x, min_y, max_x, min_y, max_x, max_y, min_x, max_y
                    ]
                ann_info = kie_dataset._parse_anno_info(annotations)
                ann_info['ori_bboxes'] = ann_info.get('ori_bboxes',
                                                      ann_info['bboxes'])
                ann_info['gt_bboxes'] = ann_info.get('gt_bboxes',
                                                     ann_info['bboxes'])
                kie_result, data = model_inference(
                    kie_model,
                    arr,
                    ann=ann_info,
                    return_data=True,
                    batch_mode=self.args.batch_mode)
                # visualize KIE results
                self.visualize_kie_output(
                    kie_model,
                    data,
                    kie_result,
                    out_file=out_file,
                    show=self.args.imshow)
                gt_bboxes = data['gt_bboxes'].data.numpy().tolist()
                labels = self.generate_kie_labels(kie_result, gt_bboxes,
                                                  kie_model.class_list)
                for i in range(len(gt_bboxes)):
                    img_e2e_res['result'][i]['label'] = labels[i][0]
                    img_e2e_res['result'][i]['label_score'] = labels[i][1]

            end2end_res.append(img_e2e_res)
        return end2end_res

    # Separate det/recog inference pipeline
    def single_inference(self, model, arrays, batch_mode, batch_size=0):

        def inference(m, a, **kwargs):
            if model == 'Tesseract_det':
                return self.tesseract_det_inference(a)
            elif model == 'Tesseract_recog':
                return self.tesseract_recog_inference(a)
            else:
                return model_inference(m, a, **kwargs)

        result = []
        if batch_mode:
            if batch_size == 0:
                result = inference(model, arrays, batch_mode=True)
            else:
                n = batch_size
                arr_chunks = [
                    arrays[i:i + n] for i in range(0, len(arrays), n)
                ]
                for chunk in arr_chunks:
                    result.extend(inference(model, chunk, batch_mode=True))
        else:
            for arr in arrays:
                result.append(inference(model, arr, batch_mode=False))
        return result

    # Arguments pre-processing function
    def _args_processing(self, args):
        # Check if the input is a list/tuple that
        # contains only np arrays or strings
        if isinstance(args.img, (list, tuple)):
            img_list = args.img
            if not all([isinstance(x, (np.ndarray, str)) for x in args.img]):
                raise AssertionError('Images must be strings or numpy arrays')

        # Create a list of the images
        if isinstance(args.img, str):
            img_path = Path(args.img)
            if img_path.is_dir():
                img_list = [str(x) for x in img_path.glob('*')]
            else:
                img_list = [str(img_path)]
        elif isinstance(args.img, np.ndarray):
            img_list = [args.img]

        # Read all image(s) in advance to reduce wasted time
        # re-reading the images for visualization output
        args.arrays = [mmcv.imread(x) for x in img_list]

        # Create a list of filenames (used for output images and result files)
        if isinstance(img_list[0], str):
            args.filenames = [str(Path(x).stem) for x in img_list]
        else:
            args.filenames = [str(x) for x in range(len(img_list))]

        # If given an output argument, create a list of output image filenames
        num_res = len(img_list)
        if args.output:
            output_path = Path(args.output)
            if output_path.is_dir():
                args.output = [
                    str(output_path / f'out_{x}.png') for x in args.filenames
                ]
            else:
                args.output = [str(args.output)]
                if args.batch_mode:
                    raise AssertionError('Output of multiple images inference'
                                         ' must be a directory')
        else:
            args.output = [None] * num_res

        # If given an export argument, create a list of
        # result filenames for each image
        if args.export:
            export_path = Path(args.export)
            args.export = [
                str(export_path / f'out_{x}.{args.export_format}')
                for x in args.filenames
            ]
        else:
            args.export = [None] * num_res

        return args


class Ocr(object):
  def __init__(self, recog_config, det_config, recog_ckpt=None, det_ckpt=None, device='cpu'):
    # recog
    if not recog_config.endswith('.py'):
      recog_config = f'{recog_config}.py'
    recog_config_name = recog_config.split('.')[0]
    model_folder = os.path.join(os.environ['HOME'], '.antgo', 'models', 'mmocr')
    if not os.path.exists(model_folder):
      os.makedirs(model_folder)

    model_file = os.path.join(model_folder, recog_config)
    if not os.path.exists(model_file):
      subprocess.check_call([sys.executable, '-m', 'mim', 'download', 'mmocr', '--config', recog_config_name, '--dest', model_folder])

    recog_config_prefix_name = recog_config_name.split('_')[0]
    if recog_ckpt is None:
      for f in os.listdir(model_folder):
        if f.startswith(recog_config_prefix_name) and f.endswith('.pth'):
          recog_ckpt = f
          break
    assert(recog_ckpt is not None)

    # det
    if not det_config.endswith('.py'):
      det_config = f'{det_config}.py'
    det_config_name = det_config.split('.')[0]
    model_file = os.path.join(model_folder, det_config)
    if not os.path.exists(model_file):
      subprocess.check_call(
        [sys.executable, '-m', 'mim', 'download', 'mmocr', '--config', det_config_name, '--dest', model_folder])

    det_config_prefix_name = det_config_name.split('_')[0]
    if det_ckpt is None:
      for f in os.listdir(model_folder):
        if f.startswith(det_config_prefix_name) and f.endswith('.pth'):
          det_ckpt = f
          break
    assert (det_ckpt is not None)

    self.model = MMOCR(recog_ckpt=recog_ckpt, recog_config=recog_config,det_ckpt=det_ckpt,det_config=det_config, device=device)

  def __call__(self, *args, **kwargs):
    result = self.model.readtext(args[0], imshow=False)
    # label
    return result[0]



# 优化器配置
optimizer = dict(type='SGD', lr=0.01,  weight_decay=5e-4, momentum=0.9, nesterov=True)
optimizer_config = dict(grad_clip=dict(max_norm=20))

# 学习率调度配置
lr_config = dict(
    policy='CosineAnnealing',
    min_lr=1e-5,
)

# 模型配置文件标准格式
# info = dict(
#     backbone=dict(
#         channels=[96,128,160],
#         shapes=[16,8,4]
#     ),
#     neck=dict(
#         channels=[],
#         shapes=[]
#     )
# )

# 模型配置
# model 字段根据具体模型进行设置
model = dict(
    type='ReviewKD',
    model=dict(
        teacher=None,   # placeholder
        student=None,   # placeholder  
    ),    
    train_cfg=dict(
        student=dict(
            channels=[96,128,160],
            shapes=[16,8,4]
        ),
        teacher=dict(
            channels=[512, 1024, 2048],
            shapes=[16,8,4]
        ),
        kd_loss_weight=1.0,
        kd_warm_up=20.0,
        align = 'backbone'
    ),
    test_cfg=None,
    init_cfg=None,
)
